PROCESSOR_NAME_KEY = "processor"
PROCESSOR_TYPE_KEY = "processor_type"
PROCESSOR_PARAMS_KEY = "processor_params"
PRE_PROCESSORS_KEY = "pre_processors"
POST_PROCESSORS_KEY = "post_processors"


class BaseProcessor(object):
    """
    Base class for upload Processor classes
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.parameters = kwargs.get(PROCESSOR_PARAMS_KEY, dict()) or dict()

    @classmethod
    def process(cls):
        raise NotImplementedError("Must be implemented by subclass")





class DerivaRestoreError(RuntimeError):
    pass


class DerivaRestoreConfigurationError(ValueError):
    pass


class DerivaRestoreAuthenticationError(Exception):
    pass


class DerivaRestoreAuthorizationError(Exception):
    pass



import io
import os
import sys
import shutil
import errno
import json
import platform
import logging
import requests
import portalocker
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests.packages.urllib3.exceptions import MaxRetryError
from collections import OrderedDict
from distutils import util as du_util

IS_PY2 = (sys.version_info[0] == 2)
IS_PY3 = (sys.version_info[0] == 3)

if IS_PY3:
    from urllib.parse import quote as _urlquote, unquote as urlunquote
    from urllib.parse import urlparse, urlsplit, urlunsplit, urljoin
    from http.cookiejar import MozillaCookieJar
else:
    from urllib import quote as _urlquote, unquote as urlunquote
    from urlparse import urlparse, urlsplit, urlunsplit, urljoin
    from cookielib import MozillaCookieJar

Kilobyte = 1024
Megabyte = Kilobyte ** 2
DEFAULT_CHUNK_SIZE = Megabyte * 10  # above the minimum 5MB chunk size for AWS S3 multipart uploads
MAX_CHUNK_SIZE = Megabyte * 100

DEFAULT_HEADERS = {}
DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser('~'), '.deriva')
DEFAULT_CREDENTIAL_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'credential.json')
DEFAULT_GLOBUS_CREDENTIAL_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'globus-credential.json')
DEFAULT_CONFIG_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'config.json')
DEFAULT_COOKIE_JAR_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'cookies.txt')
DEFAULT_SESSION_CONFIG = {
    "retry_connect": 2,
    "retry_read": 4,
    "retry_backoff_factor": 1.0,
    "retry_status_forcelist": [500, 503, 504],
    "cookie_jar": DEFAULT_COOKIE_JAR_FILE
}
OAUTH2_SCOPES_KEY = "oauth2_scopes"
DEFAULT_CONFIG = {
    "server":
    {
        "protocol": "https",
        "host": platform.uname()[1],
        "catalog_id": 1
    },
    "session": DEFAULT_SESSION_CONFIG,
    "download_processor_whitelist": [],
    OAUTH2_SCOPES_KEY: {}
}
DEFAULT_CREDENTIAL = {}


class NotModified (ValueError):
    pass


class ConcurrentUpdate (ValueError):
    pass


def urlquote(s, safe=''):
    """Quote all reserved characters according to RFC3986 unless told otherwise.

       The urllib.urlquote has a weird default which excludes '/' from
       quoting even though it is a reserved character.  We would never
       want this when encoding elements in Deriva REST API URLs, so
       this wrapper changes the default to have no declared safe
       characters.

    """
    return _urlquote(s.encode('utf-8'), safe=safe)


def urlquote_dcctx(s, safe='~{}",:'):
    """Quote for use with Deriva-Client-Context or other HTTP headers.

       Defaults to allow additional safe characters for less
       aggressive encoding of JSON content for use in an HTTP header
       value.

    """
    return urlquote(s, safe=safe)


def stob(string):
    return bool(du_util.strtobool(str(string)))


def format_exception(e):
    if not isinstance(e, Exception):
        return str(e)
    exc = "".join(("[", type(e).__name__, "] "))
    if isinstance(e, requests.HTTPError):
        resp = " - Server responded: %s" % e.response.text.strip().replace('\n', ': ')
        return "".join((exc, str(e), resp))
    return "".join((exc, str(e)))


def add_logging_level(level_name, level_num, method_name=None):
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        logging.warning('{} already defined in logging module'.format(level_name))
        return
    if hasattr(logging, method_name):
        logging.warning('{} already defined in logging module'.format(method_name))
        return
    if hasattr(logging.getLoggerClass(), method_name):
        logging.warning('{} already defined in logger class'.format(method_name))
        return

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def init_logging(level=logging.INFO,
                 log_format=None,
                 file_path=None,
                 file_mode='w',
                 capture_warnings=True):
    add_logging_level("TRACE", logging.DEBUG-5)
    logging.captureWarnings(capture_warnings)
    if log_format is None:
        log_format = "[%(asctime)s - %(levelname)s - %(filename)s:%(lineno)s:%(funcName)s()] %(message)s" \
            if level <= logging.DEBUG else "%(asctime)s - %(levelname)s - %(message)s"
    if file_path:
        logging.basicConfig(filename=file_path, filemode=file_mode, level=level, format=log_format)
    else:
        logging.basicConfig(level=level, format=log_format)


def get_new_requests_session(url=None, session_config=DEFAULT_SESSION_CONFIG):
    session = requests.session()
    retries = Retry(connect=session_config['retry_connect'],
                    read=session_config['retry_read'],
                    backoff_factor=session_config['retry_backoff_factor'],
                    status_forcelist=session_config['retry_status_forcelist'],
                    raise_on_status=True)
    if url:
        session.mount(url, HTTPAdapter(max_retries=retries))
    else:
        session.mount('http://', HTTPAdapter(max_retries=retries))
        session.mount('https://', HTTPAdapter(max_retries=retries))

    return session


def make_dirs(path, mode=0o777):
    if not os.path.isdir(path):
        try:
            os.makedirs(path, mode=mode)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise


def copy_config(src, dst):
    config_dir = os.path.dirname(dst)
    make_dirs(config_dir, mode=0o750)
    shutil.copy2(src, dst)


def write_config(config_file=DEFAULT_CONFIG_FILE, config=DEFAULT_CONFIG):
    config_dir = os.path.dirname(config_file)
    make_dirs(config_dir, mode=0o750)
    with io.open(config_file, 'w', newline='\n', encoding='utf-8') as cf:
        config_data = json.dumps(config, ensure_ascii=False, indent=2)
        if IS_PY2 and isinstance(config_data, str):
            config_data = unicode(config_data, 'utf-8')
        cf.write(config_data)
        cf.close()


def read_config(config_file=DEFAULT_CONFIG_FILE, create_default=False, default=DEFAULT_CONFIG):
    if not config_file:
        config_file = DEFAULT_CONFIG_FILE
    config = None
    if not os.path.isfile(config_file) and create_default:
        logging.info("No default configuration file found, attempting to create one at: %s" % config_file)
        try:
            write_config(config_file, default)
        except Exception as e:
            logging.warning("Unable to create configuration file %s. Using internal defaults. %s" %
                            (config_file, format_exception(e)))
            config = json.dumps(default, ensure_ascii=False)

    if not config:
        with open(config_file) as cf:
            config = cf.read()

    return json.loads(config, object_pairs_hook=OrderedDict)


def lock_file(file_path, mode, exclusive=True):
    return portalocker.Lock(file_path, mode, timeout=60,
                            flags=portalocker.LOCK_EX if exclusive else portalocker.LOCK_SH)


def write_credential(credential_file=DEFAULT_CREDENTIAL_FILE, credential=DEFAULT_CREDENTIAL):
    credential_dir = os.path.dirname(credential_file)
    make_dirs(credential_dir, mode=0o750)
    with lock_file(credential_file, mode='w', exclusive=True) as cf:
        os.chmod(credential_file, 0o600)
        credential_data = json.dumps(credential, ensure_ascii=False, indent=2)
        if IS_PY2 and isinstance(credential_data, str):
            credential_data = unicode(credential_data, 'utf-8')
        cf.write(credential_data)
        cf.flush()
        os.fsync(cf.fileno())


def read_credential(credential_file=DEFAULT_CREDENTIAL_FILE, create_default=False, default=DEFAULT_CREDENTIAL):
    if not credential_file:
        credential_file = DEFAULT_CREDENTIAL_FILE
    credential = None
    if not os.path.isfile(credential_file) and create_default:
        logging.info("No default credential file found, attempting to create one at: %s" % credential_file)
        try:
            write_credential(credential_file, default)
        except Exception as e:
            logging.warning("Unable to create credential file %s. Using internal defaults. %s" %
                            (credential_file, format_exception(e)))
            credential = json.dumps(default, ensure_ascii=False)

    if not credential:
        with lock_file(credential_file, mode='r', exclusive=False) as cf:
            credential = cf.read()

    return json.loads(credential, object_pairs_hook=OrderedDict)


def get_credential(host,
                   credential_file=DEFAULT_CREDENTIAL_FILE,
                   globus_credential_file=DEFAULT_GLOBUS_CREDENTIAL_FILE,
                   config_file=DEFAULT_CONFIG_FILE,
                   requested_scope=None,
                   force_scope_lookup=False,
                   match_scope_tag="deriva-all"):
    # load deriva credential set first
    credentials = read_credential(credential_file or DEFAULT_CREDENTIAL_FILE, create_default=True)
    creds = credentials.get(host, credentials.get(host.lower(), dict()))

    # if present, load globus credentials and merge
    globus_credentials = read_credential(globus_credential_file or DEFAULT_GLOBUS_CREDENTIAL_FILE, create_default=True)
    if globus_credentials:
        scopes = get_oauth_scopes_for_host(host,
                                           config_file=config_file,
                                           force_refresh=force_scope_lookup,
                                           warn_on_discovery_failure=True if not creds else False)
        for resource, g_creds in globus_credentials.items():
            # 1. look for the explicitly requested scope in the token store, if specified
            if requested_scope is not None and g_creds["scope"] == requested_scope:
                creds["bearer-token"] = g_creds["access_token"]
                break

            # 2. try to determine the scope to use based on host-to-scope(s) mappings
            if scopes:
                for k, v in scopes.items():
                    if v == g_creds["scope"]:
                        if match_scope_tag is not None and match_scope_tag != k:
                            continue
                        creds["bearer-token"] = g_creds["access_token"]
                        break
                if creds.get("bearer-token"):
                    break
            else:
                # 3. if we did not find any host-to-scope(s) mappings in the config file, just fall back and
                # try to find a token by matching the hostname to the resource server name
                if host.lower() == resource.lower():
                    creds["bearer-token"] = g_creds["access_token"]
                    break

    return creds or None


def get_oauth_scopes_for_host(host,
                              config_file=DEFAULT_CONFIG_FILE,
                              force_refresh=False,
                              warn_on_discovery_failure=False):
    config = read_config(config_file or DEFAULT_CONFIG_FILE, create_default=True)
    required_scopes = config.get(OAUTH2_SCOPES_KEY)
    result = dict()
    upr = urlparse(host)
    if upr.scheme and upr.netloc:
        url = urljoin(host, "/authn/discovery")
        host = upr.hostname
    else:
        url = "https://%s/authn/discovery" % host
    # determine the scope to use based on host-to-scope(s) mappings in the config file
    if required_scopes:
        for hostname, scopes in required_scopes.items():
            if host.lower() == hostname.lower():
                result = scopes
                break
    if not result or force_refresh:
        session = get_new_requests_session(session_config=DEFAULT_SESSION_CONFIG)
        try:
            r = session.get(url, headers=DEFAULT_HEADERS)
            r.raise_for_status()
            result = r.json().get(OAUTH2_SCOPES_KEY)
            if result:
                if not config.get(OAUTH2_SCOPES_KEY):
                    config[OAUTH2_SCOPES_KEY] = {}
                config[OAUTH2_SCOPES_KEY].update({host: result})
                write_config(config_file or DEFAULT_CONFIG_FILE, config=config)
        except Exception as e:
            msg = "Unable to discover and/or update the \"%s\" mappings from [%s]. Requests to this host that " \
                  "use OAuth2 scope-specific bearer tokens may fail or provide only limited access. %s" % \
                  (OAUTH2_SCOPES_KEY, url, format_exception(e))
            if warn_on_discovery_failure:
                logging.warning(msg)
            else:
                logging.debug(msg)
        finally:
            session.close()
    return result


def format_credential(token=None, oauth2_token=None, username=None, password=None):
    if username and password:
        return {"username": username, "password": password}
    credential = dict()
    if token:
        credential.update({"cookie": "webauthn=%s" % token})
    if oauth2_token:
        credential.update({"bearer-token": "%s" % oauth2_token})
    if not credential:
        raise ValueError(
            "Missing required argument(s): a supported authentication token or a username/password must be provided.")

    return credential


def bootstrap():
    init_logging()
    read_config(create_default=True)
    read_credential(create_default=True)


def load_cookies_from_file(cookie_file=None):
    if not cookie_file:
        cookie_file = DEFAULT_SESSION_CONFIG["cookie_jar"]
    cookies = MozillaCookieJar()
    # Load and return saved cookies if existing
    if os.path.isfile(cookie_file):
        try:
            cookies.load(cookie_file, ignore_discard=True, ignore_expires=True)
            return cookies
        except Exception as e:
            logging.warning(format_exception(e))

    # Create new empty cookie file otherwise
    cookies.save(cookie_file, ignore_discard=True, ignore_expires=True)
    os.chmod(cookie_file, 0o600)

    return cookies


def resource_path(relative_path, default=os.path.abspath(".")):
    # required to find bundled data at runtime in Pyinstaller single-file exe mode
    if getattr(sys, 'frozen', False):
        return os.path.join(getattr(sys, '_MEIPASS', '.'), relative_path)
    if default is None:
        return relative_path
    return os.path.join(default, relative_path)


def get_transfer_summary(total_bytes, elapsed_time):
    total_secs = elapsed_time.total_seconds()
    transferred = \
        float(total_bytes) / float(Kilobyte) if total_bytes < Megabyte else float(total_bytes) / float(Megabyte)
    throughput = str(" at %.2f MB/second" % (transferred / total_secs)) if (total_secs >= 1) else ""
    elapsed = str("Elapsed time: %s." % elapsed_time) if (total_secs > 0) else ""
    summary = "%.2f %s transferred%s. %s" % \
              (transferred, "KB" if total_bytes < Megabyte else "MB", throughput, elapsed)
    return summary


def topo_sorted(depmap):
    """Return list of items topologically sorted.

       depmap: { item: [required_item, ...], ... }

    Raises ValueError if a required_item cannot be satisfied in any order.

    The per-item required_item iterables must allow revisiting on
    multiple iterations.

    """
    ordered = [ item for item, requires in depmap.items() if not requires ]
    depmap = { item: set(requires) for item, requires in depmap.items() if requires }
    satisfied = set(ordered)
    while depmap:
        additions = []
        for item, requires in list(depmap.items()):
            if requires.issubset(satisfied):
                additions.append(item)
                satisfied.add(item)
                del depmap[item]
        if not additions:
            raise ValueError(("unsatisfiable", depmap))
        ordered.extend(additions)
        additions = []
    return ordered


class AttrDict (dict):
    """Dictionary with optional attribute-based lookup.

       For keys that are valid attributes, self.key is equivalent to
       self[key].
    """
    def __getattr__(self, a):
        try:
            return self[a]
        except KeyError as e:
            raise AttributeError(str(e))

    def __setattr__(self, a, v):
        self[a] = v

    def update(self, d):
        dict.update(self, d)


# convenient enumeration of common annotation tags
tag = AttrDict({
    'display':            'tag:misd.isi.edu,2015:display',
    'table_alternatives': 'tag:isrd.isi.edu,2016:table-alternatives',
    'column_display':     'tag:isrd.isi.edu,2016:column-display',
    'key_display':        'tag:isrd.isi.edu,2017:key-display',
    'foreign_key':        'tag:isrd.isi.edu,2016:foreign-key',
    'generated':          'tag:isrd.isi.edu,2016:generated',
    'immutable':          'tag:isrd.isi.edu,2016:immutable',
    'non_deletable':      'tag:isrd.isi.edu,2016:non-deletable',
    'app_links':          'tag:isrd.isi.edu,2016:app-links',
    'table_display':      'tag:isrd.isi.edu,2016:table-display',
    'visible_columns':    'tag:isrd.isi.edu,2016:visible-columns',
    'visible_foreign_keys': 'tag:isrd.isi.edu,2016:visible-foreign-keys',
    'export':             'tag:isrd.isi.edu,2016:export',
    'export_2019':        'tag:isrd.isi.edu,2019:export',
    'asset':              'tag:isrd.isi.edu,2017:asset',
    'citation':           'tag:isrd.isi.edu,2018:citation',
    'required':           'tag:isrd.isi.edu,2018:required',
    'indexing_preferences': 'tag:isrd.isi.edu,2018:indexing-preferences',
    'bulk_upload':        'tag:isrd.isi.edu,2017:bulk-upload',
    'chaise_config':      'tag:isrd.isi.edu,2019:chaise-config',
    'source_definitions': 'tag:isrd.isi.edu,2019:source-definitions'
})



import io
import os
import sys
import copy
import json
import time
import logging
import datetime
import platform
from collections import OrderedDict
from bdbag import bdbag_api as bdb
from deriva.core import get_credential, format_credential, urlquote, format_exception, __version__ as VERSION
from deriva.core.utils.version_utils import get_installed_version
from deriva.core.ermrest_model import Model
from deriva.core.deriva_server import DerivaServer
from deriva.core.ermrest_catalog import ErmrestCatalog, _clone_state_url as CLONE_STATE_URL
from deriva.core.hatrac_store import HatracStore
from deriva.transfer import DerivaUpload, DerivaUploadError, DerivaUploadConfigurationError, GenericUploader
from deriva.transfer.restore import DerivaRestoreError, DerivaRestoreConfigurationError, \
    DerivaRestoreAuthenticationError, DerivaRestoreAuthorizationError


class DerivaRestore:
    """
    Restore a DERIVA catalog from a bag archive or directory.
    Core restore logic re-purposed from ErmrestCatalog.clone_catalog().
    """

    RESTORE_STATE_URL = "tag:isrd.isi.edu,2019:restore-status"
    BASE_DATA_INPUT_PATH = os.path.join("records", "{}", "{}.json")
    BASE_ASSETS_INPUT_PATH = "assets"

    def __init__(self, *args, **kwargs):

        self.server_args = args[0]
        self.hostname = None
        self.dst_catalog = None
        self.cancelled = False
        self.input_path = kwargs.get("input_path")
        self.exclude_schemas = kwargs.get("exclude_schemas", list())
        self.restore_data = not kwargs.get("no_data", False)
        self.data_chunk_size = kwargs.get("data_chunk_size", 10000)
        self.restore_annotations = not kwargs.get("no_annotations", False)
        self.restore_policy = not kwargs.get("no_policy", False)
        self.restore_assets = not kwargs.get("no_assets", False)
        self.strict_bag_validation = not kwargs.get("weak_bag_validation", True)
        self.no_bag_materialize = kwargs.get("no_bag_materialize", False)
        self.upload_config = kwargs.get("asset_config")
        self.truncate_after = True
        self.envars = kwargs.get("envars", dict())
        self.config = kwargs.get("config")
        self.credentials = kwargs.get("credentials", dict())
        config_file = kwargs.get("config_file")
        credential_file = kwargs.get("credential_file")

        info = "%s v%s [Python %s, %s]" % (
            self.__class__.__name__, get_installed_version(VERSION),
            platform.python_version(), platform.platform(aliased=True))
        logging.info("Initializing: %s" % info)

        if not self.server_args:
            raise DerivaRestoreConfigurationError("Target server not specified!")

        # server variable initialization
        self.hostname = self.server_args.get('host', '')
        if not self.hostname:
            raise DerivaRestoreConfigurationError("Host not specified!")
        protocol = self.server_args.get('protocol', 'https')
        self.server_url = protocol + "://" + self.hostname
        self.catalog_id = self.server_args.get("catalog_id",)
        self.session_config = self.server_args.get('session')

        # credential initialization
        token = kwargs.get("token")
        oauth2_token = kwargs.get("oauth2_token")
        username = kwargs.get("username")
        password = kwargs.get("password")
        if token or oauth2_token or (username and password):
            self.credentials = format_credential(token=token,
                                                 oauth2_token=oauth2_token,
                                                 username=username,
                                                 password=password)
        else:
            self.credentials = get_credential(self.hostname, credential_file)

        # destination catalog initialization
        self.server = DerivaServer(protocol,
                                   self.hostname,
                                   self.credentials,
                                   caching=True,
                                   session_config=self.session_config)
        self.server.dcctx["cid"] = kwargs.get("dcctx_cid", "api/" + self.__class__.__name__)
        # process config file
        if config_file:
            try:
                self.config = read_config(config_file)
            except Exception as e:
                raise DerivaRestoreConfigurationError(e)

    def set_config(self, config):
        self.config = config

    def set_credentials(self, credentials):
        self.catalog.set_credentials(credentials, self.hostname)
        self.store.set_credentials(credentials, self.hostname)
        self.credentials = credentials

    def prune_parts(self, dest):
        if not self.restore_annotations and 'annotations' in dest:
            del dest['annotations']
        if not self.restore_policy:
            if 'acls' in dest:
                del dest['acls']
            if 'acl_bindings' in dest:
                del dest['acl_bindings']
        return dest

    def copy_sdef(self, schema):
        """Copy schema definition structure with conditional parts for cloning."""
        dest = self.prune_parts(schema.prejson())
        if 'tables' in dest:
            del dest['tables']
        return dest

    def copy_tdef_core(self, table):
        """Copy table definition structure with conditional parts excluding fkeys."""
        dest = self.prune_parts(table.prejson())
        dest['column_definitions'] = [self.prune_parts(column) for column in dest['column_definitions']]
        dest['keys'] = [self.prune_parts(column) for column in dest.get('keys', [])]
        if 'foreign_keys' in dest:
            del dest['foreign_keys']
        if 'annotations' not in dest:
            dest['annotations'] = {}
        dest['annotations'][self.RESTORE_STATE_URL] = 1 if self.restore_data else None
        return dest

    def copy_tdef_fkeys(self, table):
        """Copy table fkeys structure."""

        def check(fkdef):
            for fkc in fkdef['referenced_columns']:
                if fkc['schema_name'] == 'public' \
                        and fkc['table_name'] in {'ERMrest_Client', 'ERMrest_Group'} \
                        and fkc['column_name'] == 'RID':
                    raise DerivaRestoreError(
                        "Cannot restore catalog with foreign key reference to "
                        "%(schema_name)s:%(table_name)s:%(column_name)s" % fkc)
            return fkdef

        return [self.prune_parts(check(dest)) for dest in table.prejson().get('foreign_keys', [])]

    def copy_cdef(self, column):
        """Copy column definition with conditional parts."""
        return column.table.schema.name, column.table.name, self.prune_parts(column.prejson())

    @staticmethod
    def check_column_compatibility(src, dst):
        """Check compatibility of source and destination column definitions."""

        def error(fieldname, sv, dv):
            return DerivaRestoreError("Source/dest column %s mismatch %s != %s for %s:%s:%s" % (
                fieldname,
                sv, dv,
                src.sname, src.tname, src.name
            ))

        if src.type.typename != dst.type.typename:
            raise error("type", src.type.typename, dst.type.typename)
        if src.nullok != dst.nullok:
            raise error("nullok", src.nullok, dst.nullok)
        if src.default != dst.default:
            raise error("default", src.default, dst.default)

    def copy_kdef(self, key):
        return key.table.schema.name, key.table.name, self.prune_parts(key.prejson())

    def get_table_path(self, sname, tname, is_bag):
        return os.path.abspath(
            os.path.join(self.input_path, "data" if is_bag else "", self.BASE_DATA_INPUT_PATH.format(sname, tname)))

    def load_json_file(self, file_path):
        with io.open(file_path, 'r', encoding='UTF-8') as file_data:
            return json.load(file_data, object_pairs_hook=OrderedDict)

    def open_json_stream_file(self, table_path):
        """
        Open a JSON-Stream file for reading, caller is responsible for closing.
        """
        table_data = io.open(table_path, 'r', encoding='UTF-8')
        line = table_data.readline().strip()
        table_data.seek(0)
        if line.startswith('{') and line.endswith('}'):
            return table_data
        else:
            table_data.close()
            raise DerivaRestoreError(
                "Input file %s does not appear to be in the required json-stream format." % table_path)

    def get_json_recordset(self, data, chunk_size, after=None, after_column='RID'):
        chunk = list()
        found = False
        for line in data:
            if isinstance(line, dict):
                row = line
            else:
                row = json.loads(line, object_pairs_hook=OrderedDict)
            if after and not found:
                if after == row[after_column]:
                    found = True
                continue
            chunk.append(row)
            if len(chunk) == chunk_size:
                yield chunk
                chunk = list()
        if chunk:
            yield chunk

    def restore(self, **kwargs):
        """
        Perform the catalog restore operation. The restore process is broken up into six phases:

        1. Pre-process the input path.
            - If the input path is a file, it is assumed that it is a compressed archive file that can be extracted
            into an input directory via a supported codec: `tar`,`tgz`,`bz2`, or `zip`.
            - If the input directory is a valid _bag_ directory structure, the bag will be materialized.
        2. The catalog schema will be restored first. The schema is restored from a ERMRest JSON schema document file.
            The schema document file must be named `catalog-schema.json` and must appear at the root of the input
            directory. The restore process can be configured to exclude the restoration of an enumerated set both
            schema and tables.
        3. The catalog table data will be restored, if present. The table date restoration process is resilient to
            interruption and may be restarted. However, if the catalog schema or data is mutated outside of the scope of
            the restore function in-between such restarts, the restored catalog's consistency cannot be guaranteed.
            The restore process can be configured to exclude the restoration of table data for a set of tables.
        4. The catalog foreign keys will be restored.
        5. The catalog assets will be restored, if present.
        6. On success, the restore state marker annotations will be deleted and the catalog history will be truncated.

        :param kwargs:
        :return:
        """
        success = True
        start = datetime.datetime.now()

        # pre-process input
        logging.info("Processing input path: %s" % self.input_path)
        is_file, is_dir, is_uri = bdb.inspect_path(self.input_path)
        if not (is_file or is_dir or is_uri):
            raise DerivaRestoreError("Invalid input path [%s]. If the specified input path refers to a locally mounted "
                                     "file or directory, it does not exist or cannot be accessed. If the specified "
                                     "path is a URI, the scheme component of the URI could not be determined." %
                                     self.input_path)
        if is_file or is_dir:
            self.input_path = os.path.abspath(self.input_path)
        if is_file:
            logging.info("The input path [%s] is a file. Assuming input file is a directory archive and extracting..." %
                         self.input_path)
            self.input_path = bdb.extract_bag(self.input_path)

        try:
            if not self.no_bag_materialize:
                self.input_path = bdb.materialize(self.input_path)
        except bdb.bdbagit.BagValidationError as e:
            if self.strict_bag_validation:
                raise DerivaRestoreError(format_exception(e))
            else:
                logging.warning("Input bag validation failed and strict validation mode is disabled. %s" %
                                format_exception(e))
        is_bag = bdb.is_bag(self.input_path)

        src_schema_file = os.path.abspath(
            os.path.join(self.input_path, "data" if is_bag else "", "catalog-schema.json"))
        # the src_catalog_stub created below will never be "connected" in any kind of network sense,
        # but we need an instance of ErmrestCatalog in order to get a working Model from the schema file.
        src_catalog_stub = ErmrestCatalog("file", src_schema_file, "1")
        src_model = Model.fromfile(src_catalog_stub, src_schema_file)

        # initialize/connect to destination catalog
        if not self.catalog_id:
            self.catalog_id = self.server.create_ermrest_catalog().catalog_id
            self.server_args["catalog_id"] = self.catalog_id
            logging.info("Created new target catalog with ID: %s" % self.catalog_id)
        self.dst_catalog = self.server.connect_ermrest(self.catalog_id)

        # init dcctx cid to a default
        self.dst_catalog.dcctx['cid'] = self.__class__.__name__

        # build up the model content we will copy to destination
        dst_model = self.dst_catalog.getCatalogModel()

        logging.info("Restoring %s to catalog: %s" % (self.input_path, self.dst_catalog.get_server_uri()))
        # set top-level config right away and find fatal usage errors...
        if self.restore_policy:
            logging.info("Restoring top-level catalog ACLs...")
            if not src_model.acls:
                logging.info("Source schema does not contain any ACLs.")
            else:
                src_model.acls.owner.extend(dst_model.acls.owner)
                self.dst_catalog.put('/acl', json=src_model.acls)

        if self.restore_annotations:
            logging.info("Restoring top-level catalog annotations...")
            self.dst_catalog.put('/annotation', json=src_model.annotations)

        # build up the model content we will copy to destination
        dst_model = self.dst_catalog.getCatalogModel()

        new_model = []
        new_columns = []  # ERMrest does not currently allow bulk column creation
        new_keys = []  # ERMrest does not currently allow bulk key creation
        restore_states = {}
        fkeys_deferred = {}
        exclude_schemas = [] if self.exclude_schemas is None else self.exclude_schemas

        try:
            for sname, schema in src_model.schemas.items():
                if sname in exclude_schemas:
                    continue
                if sname not in dst_model.schemas:
                    new_model.append(self.copy_sdef(schema))

                for tname, table in schema.tables.items():
                    if table.kind != 'table':
                        logging.warning('Skipping restore of %s %s:%s' % (table.kind, sname, tname))
                        continue

                    if 'RID' not in table.column_definitions.elements:
                        raise DerivaRestoreError(
                            "Source table %s.%s lacks system-columns and cannot be restored." % (sname, tname))

                    # make sure the source table is pruned of any existing restore state markers
                    if table.annotations.get(CLONE_STATE_URL) is not None:
                        del table.annotations[CLONE_STATE_URL]
                    if table.annotations.get(self.RESTORE_STATE_URL) is not None:
                        del table.annotations[self.RESTORE_STATE_URL]

                    if sname not in dst_model.schemas or tname not in dst_model.schemas[sname].tables:
                        new_model.append(self.copy_tdef_core(table))
                        restore_states[(sname, tname)] = 1 if self.restore_data else None
                        fkeys_deferred[(sname, tname)] = self.copy_tdef_fkeys(table)
                    else:
                        src_columns = {c.name: c for c in table.column_definitions}
                        dst_columns = {c.name: c for c in dst_model.schemas[sname].tables[tname].column_definitions}

                        for cname in src_columns:
                            if cname not in dst_columns:
                                new_columns.append(self.copy_cdef(src_columns[cname]))
                            else:
                                self.check_column_compatibility(src_columns[cname], dst_columns[cname])

                        for cname in dst_columns:
                            if cname not in src_columns:
                                raise DerivaRestoreError(
                                    "Destination column %s.%s.%s does not exist in source catalog." %
                                    (sname, tname, cname))

                        src_keys = {tuple(sorted(c.name for c in key.unique_columns)): key for key in table.keys}
                        dst_keys = {tuple(sorted(c.name for c in key.unique_columns)): key for key in
                                    dst_model.schemas[sname].tables[tname].keys}

                        for utuple in src_keys:
                            if utuple not in dst_keys:
                                new_keys.append(self.copy_kdef(src_keys[utuple]))

                        for utuple in dst_keys:
                            if utuple not in src_keys:
                                raise DerivaRestoreError("Destination key %s.%s(%s) does not exist in source catalog."
                                                         % (sname, tname, ', '.join(utuple)))

                        restore_states[(sname, tname)] = \
                            dst_model.schemas[sname].tables[tname].annotations.get(self.RESTORE_STATE_URL)
                        if dst_model.schemas[sname].tables[tname].foreign_keys:
                            # assume that presence of any destination foreign keys means we already completed
                            if self.restore_assets:
                                self.upload_assets()
                            return
                        else:
                            fkeys_deferred[(sname, tname)] = self.copy_tdef_fkeys(table)

            # apply the stage 1 model to the destination in bulk
            logging.info("Restoring catalog schema...")
            if new_model:
                self.dst_catalog.post("/schema", json=new_model).raise_for_status()

            for sname, tname, cdef in new_columns:
                self.dst_catalog.post("/schema/%s/table/%s/column" % (urlquote(sname), urlquote(tname)),
                                      json=cdef).raise_for_status()

            for sname, tname, kdef in new_keys:
                self.dst_catalog.post("/schema/%s/table/%s/key" % (urlquote(sname), urlquote(tname)),
                                      json=kdef).raise_for_status()

            # copy data in stage 2
            if self.restore_data:
                logging.info("Restoring catalog data...")
                for sname, tname in restore_states.keys():
                    tname_uri = "%s:%s" % (urlquote(sname), urlquote(tname))
                    if restore_states[(sname, tname)] == 1:
                        # determine current position in (partial?) copy
                        row = self.dst_catalog.get("/entity/%s@sort(RID::desc::)?limit=1" % tname_uri).json()
                        if row:
                            last = row[0]['RID']
                            logging.info("Existing data detected in table [%s] -- will attempt partial restore of "
                                         "remaining records following last known RID: %s" % (tname_uri, last))
                        else:
                            last = None

                        table_path = self.get_table_path(sname, tname, is_bag)
                        if not os.path.isfile(table_path):
                            logging.warning("Restoration of table data [%s] incomplete. File not found: %s" %
                                            (("%s:%s" % (sname, tname)), table_path))
                            continue
                        table = self.get_json_recordset(self.open_json_stream_file(table_path),
                                                        self.data_chunk_size, after=last)

                        total = 0
                        table_success = True
                        try:
                            for chunk in table:
                                if chunk:
                                    self.dst_catalog.post("/entity/%s?nondefaults=RID,RCT,RCB" % tname_uri, json=chunk)
                                    total += len(chunk)
                                else:
                                    break
                        except:
                            table_success = False
                        finally:
                            table.close()
                            if table_success:
                                logging.info("Restoration of table data [%s] successful. %s rows restored." %
                                             (tname_uri, total))
                            else:
                                logging.warning("Restoration of table data [%s] failed. %s rows restored." %
                                                (tname_uri, total))

                        # record our progress on catalog in case we fail part way through
                        self.dst_catalog.put(
                            "/schema/%s/table/%s/annotation/%s" % (
                                urlquote(sname),
                                urlquote(tname),
                                urlquote(self.RESTORE_STATE_URL),
                            ),
                            json=2
                        )
                    elif restore_states[(sname, tname)] is None and (sname, tname) in {
                        ('public', 'ERMrest_Client'),
                        ('public', 'ERMrest_Group'),
                    }:
                        # special sync behavior for magic ermrest tables
                        # HACK: these are assumed small enough to join via local merge of arrays
                        want = sorted(self.load_json_file(self.get_table_path(sname, tname, is_bag)),
                                      key=lambda r: r['ID'])
                        have = sorted(self.dst_catalog.get("/entity/%s?limit=none" % tname_uri).json(),
                                      key=lambda r: r['ID'])
                        create = []
                        update = []

                        pos_want = 0
                        pos_have = 0
                        while pos_want < len(want):
                            while pos_have < len(have) and have[pos_have]['ID'] < want[pos_want]['ID']:
                                # dst-only rows will be retained as is
                                pos_have += 1
                            if pos_have >= len(have) or have[pos_have]['ID'] > want[pos_want]['ID']:
                                # src-only rows will be inserted
                                create.append(want[pos_want])
                                pos_want += 1
                            else:
                                # overlapping rows will be updated
                                update.append(want[pos_want])
                                pos_want += 1

                        self.dst_catalog.post("/entity/%s?nondefaults=RCT,RCB" % tname_uri, json=create)
                        self.dst_catalog.put(
                            "/attributegroup/%s/ID;%s" % (
                                tname_uri,
                                ",".join([
                                    urlquote(c.name)
                                    for c in src_model.schemas[sname].tables[tname].column_definitions
                                    if c.name not in {'RID', 'RMT', 'RMB', 'ID'}
                                ])
                            ),
                            json=update
                        )

                        # record our progress on catalog in case we fail part way through
                        self.dst_catalog.put(
                            "/schema/%s/table/%s/annotation/%s" % (
                                urlquote(sname),
                                urlquote(tname),
                                urlquote(self.RESTORE_STATE_URL),
                            ),
                            json=2
                        )

            # apply stage 2 model in bulk only... we won't get here unless preceding succeeded
            logging.info("Restoring foreign keys...")
            new_fkeys = []
            for fkeys in fkeys_deferred.values():
                new_fkeys.extend(fkeys)

            # restore fkeys
            if new_fkeys:
                self.dst_catalog.post("/schema", json=new_fkeys)

            # restore assets
            if self.restore_assets:
                self.upload_assets()

            # cleanup
            self.cleanup_restored_catalog()
        except:
            success = False
            raise
        finally:
            elapsed_time = datetime.datetime.now() - start
            total_secs = elapsed_time.total_seconds()
            elapsed = time.strftime('%H:%M:%S', time.gmtime(total_secs))
            logging.info("Restore of catalog %s %s. %s" % (self.dst_catalog.get_server_uri(),
                                                           "completed successfully" if success else "failed",
                                                           ("Elapsed time: %s" % elapsed) if (total_secs > 0) else ""))

    def cleanup_restored_catalog(self):
        # cleanup restore state markers
        logging.info("Cleaning up restore state...")
        dst_model = self.dst_catalog.getCatalogModel()
        for sname, schema in dst_model.schemas.items():
            for tname, table in schema.tables.items():
                annotation_uri = "/schema/%s/table/%s/annotation/%s" % (
                    urlquote(sname),
                    urlquote(tname),
                    urlquote(self.RESTORE_STATE_URL)
                )
                try:
                    self.dst_catalog.delete(annotation_uri)
                except Exception as e:
                    logging.warning("Unable to cleanup restore state marker annotation %s: %s" %
                                    (annotation_uri, format_exception(e)))
                    continue

        # truncate restore history
        if self.truncate_after:
            logging.info("Truncating restore history...")
            snaptime = self.dst_catalog.get("/").json()["snaptime"]
            self.dst_catalog.delete("/history/,%s" % urlquote(snaptime))

    def upload_assets(self):
        asset_dir = os.path.join(self.input_path, self.BASE_ASSETS_INPUT_PATH)
        if not os.path.isdir(asset_dir):
            logging.debug("No asset directory found. Will not attempt to upload file assets.")
            return

        logging.info("Restoring file assets...")
        uploader = GenericUploader(config_file=self.upload_config, server=self.server_args)
        uploader.setCredentials(self.credentials)
        uploader.setConfig(self.upload_config)
        uploader.scanDirectory(asset_dir, abort_on_invalid_input=False, purge_state=False)
        uploader.uploadFiles(file_callback=uploader.defaultFileCallback)
        uploader.cleanup()



import os
import sys
import stat
import json
import logging
import platform
import traceback
import importlib
from pprint import pprint
from requests.exceptions import HTTPError, ConnectionError
from bdbag.fetch.auth import keychain as bdbkc
from deriva.core import __version__ as VERSION, DEFAULT_CONFIG_PATH, DEFAULT_GLOBUS_CREDENTIAL_FILE, urlparse, urljoin,\
    read_config, format_exception, BaseCLI, get_oauth_scopes_for_host
from deriva.core.utils import eprint

GLOBUS_SDK = None
NATIVE_LOGIN = None
NATIVE_APP_CLIENT_ID = "8ef15ba9-2b4a-469c-a163-7fd910c9d111"
LEGACY_GROUPS_SCOPE_ID = "69a73d8f-cd45-4e37-bb3b-43678424aeb7"
LEGACY_GROUPS_SCOPE_NAME = "urn:globus:auth:scope:nexus.api.globus.org:groups"
GROUPS_SCOPE_NAME = "urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships"
CLIENT_CRED_FILE = '/home/secrets/oauth2/client_secret_globus.json'
DEFAULT_SCOPES = ["openid"]


class UsageException(ValueError):
    """Usage exception.
    """

    def __init__(self, message):
        """Initializes the exception.
        """
        super(UsageException, self).__init__(message)


class DependencyError(ImportError):
    """Dependency exception.
    """

    def __init__(self, message):
        """Initializes the exception.
        """
        super(DependencyError, self).__init__(message)


class GlobusAuthAPIException(Exception):
    pass


GlobusAuthAPIError = GlobusAuthAPIException


class GlobusAuthUtil:
    def __init__(self, **kwargs):
        client_id = kwargs.get("client_id")
        client_secret = kwargs.get("client_secret")
        if not (client_id and client_secret):
            cred_file = kwargs.get("credential_file", CLIENT_CRED_FILE)
            creds = read_config(cred_file)
            if creds:
                client = creds.get("web")
                if client:
                    client_id = client.get('client_id')
                    client_secret = client.get('client_secret')
        try:
            global GLOBUS_SDK, GlobusAuthAPIError
            GLOBUS_SDK = importlib.import_module("globus_sdk")
            GlobusAuthAPIError = GLOBUS_SDK.AuthAPIError
        except Exception as e:
            raise DependencyError("Unable to load a required module: %s" % format_exception(e))

        if not (client_id and client_secret):
            logging.warning("Client ID and secret not specified and/or could not be determined.")
        self.client = GLOBUS_SDK.ConfidentialAppAuthClient(client_id, client_secret)
        self.client_id = client_id

    @staticmethod
    def from_json(obj):
        # if "obj" is a dict, use it. Otherwise, see if it's a JSON string or a file containing json.
        if isinstance(obj, dict):
            return obj
        try:
            result = json.loads(obj)
        except ValueError:
            # if this fails, we have nothing left to try, so don't bother catching errors
            f = open(obj, 'r')
            result = json.load(f)
            f.close()
        return result

    def list_all_scopes(self):
        r = self.client.get("/v2/api/scopes")
        return r.data

    def list_scope(self, scope):
        r = self.client.get("/v2/api/scopes/{s}".format(s=scope))
        return r.data

    def create_scope(self, scope):
        if not scope:
            raise UsageException("A supported scope argument is required.")

        r = self.client.post("/v2/api/clients/{client_id}/scopes".format(client_id=self.client_id),
                             json_body=self.from_json(scope))
        return r.data

    def update_scope(self, scope_id, scope):
        if not (scope_id and scope):
            raise UsageException("The scope_id and scope arguments are required.")

        r = self.client.put("/v2/api/scopes/{scope_id}".format(scope_id=scope_id),
                            json_body=self.from_json(scope))
        return r.data

    def add_fqdn_to_client(self, fqdn):
        if not fqdn:
            raise UsageException("A fqdn (fully qualified domain name) argument is required.")
        r = self.client.post('/v2/api/clients/{client_id}/fqdns'.format(client_id=self.client_id),
                             json_body={'fqdn': fqdn})
        return r.data

    def get_client_for_fqdn(self, fqdn):
        if not fqdn:
            raise UsageException("A fqdn (fully qualified domain name) argument is required.")
        r = self.client.get('/v2/api/clients?fqdn={fqdn}'.format(fqdn=fqdn))
        return r.data

    def get_client(self, client_id):
        r = self.client.get('/v2/api/clients/{client_id}'.format(client_id=client_id if client_id else self.client_id))
        return r.data

    def get_clients(self):
        r = self.client.get('/v2/api/clients')
        return r.data

    def verify_access_token(self, token):
        if not token:
            raise UsageException("A token argument is required.")
        r = self.client.oauth2_validate_token(token)
        return r.data

    def introspect_access_token(self, token):
        if not token:
            raise UsageException("A token argument is required.")
        r = self.client.oauth2_token_introspect(token)
        return r.data

    def get_dependent_access_tokens(self, token, refresh=False):
        if not token:
            raise UsageException("A token argument is required.")
        additional_params = {"access_type": "offline"} if refresh else None
        r = self.client.oauth2_get_dependent_tokens(token, additional_params)
        return r.data

    def new_client(self, client):
        r = self.client.post("/v2/api/clients",
                             json_body=self.from_json(client))
        return r.data

    def create_client(self,
                      name,
                      redirect_uris,
                      public=False,
                      visibility="private",
                      project=None,
                      required_idp=None,
                      preselect_idp=None,
                      terms_of_service=None,
                      privacy_policy=None):
        if not (name and redirect_uris):
            raise UsageException("The name and redirect_uris arguments are required.")
        client = {
            "client": {
                "name": name,
                "public_client": public,
                "redirect_uris": redirect_uris,
                "visibility": visibility if (visibility == "private" or visibility == "public") else "private"
            }
        }
        if project:
            client["client"].update({"project": project})
        if required_idp:
            client["client"].update({"required_idp": required_idp})
        if preselect_idp:
            client["client"].update({"preselect_idp": preselect_idp})
        if terms_of_service or privacy_policy:
            client["client"].update({"links": {"terms_of_service": terms_of_service,
                                               "privacy_policy": privacy_policy}})
        return self.new_client(client)

    def update_client(self, client, client_id=None):
        if not client:
            raise UsageException("A client argument is required.")

        r = self.client.put("/v2/api/clients/{client_id}".format(client_id=client_id if client_id else self.client_id),
                            json_body=self.from_json(client))
        return r.data

    def delete_client(self, client_id):
        if not client_id:
            raise UsageException("A client argument is required.")

        r = self.client.delete("/v2/api/clients/{client_id}".format(client_id=client_id))
        return r.data

    def add_redirect_uris(self, redirect_uris):
        if not redirect_uris:
            raise UsageException("A redirect_uris argument is required.")
        d = {
            "client": {
                "redirect_uris": redirect_uris
            }
        }
        return self.update_client(d)

    def get_my_client(self):
        r = self.client.get('/v2/api/clients/{client_id}'.format(client_id=self.client_id))
        return r.data

    def get_scopes_by_name(self, scope_name):
        if not scope_name:
            raise UsageException("A scope_name argument is required.")
        scopes = self.client.get('/v2/api/scopes?scope_strings={sname}'.format(sname=scope_name))
        if scopes is None:
            return None
        else:
            return scopes.get("scopes")

    def get_scopes_by_id(self, id_string):
        if not id_string:
            raise UsageException("A id_string argument is required.")
        scopes = self.client.get('/v2/api/scopes?ids={ids}'.format(ids=id_string))
        if scopes is None:
            return None
        else:
            return scopes.get("scopes")

    def my_scope_ids(self):
        c = self.client.get('/v2/api/clients/{client_id}'.format(client_id=self.client_id))
        me = c.get("client")
        if me is None or me.get('scopes') is None:
            return []
        else:
            return me.get('scopes')

    def my_scope_names(self):
        scope_names = []
        scope_ids = self.my_scope_ids()
        if scope_ids is not None:
            ids = ",".join(scope_ids)
            print(str(ids))
            scopes = self.get_scopes_by_id(ids)
            for scope in scopes:
                scope_names.append(scope.get('scope_string'))
        return scope_names

    def get_grant_types(self):
        grant_types = None
        c = self.client.get('/v2/api/clients/{client_id}'.format(client_id=self.client_id))
        me = c.get("client")
        if me is not None:
            grant_types = me.get('grant_types')
        return grant_types

    def add_scopes(self, new_scopes, by_id=False, overwrite=False):
        if not new_scopes:
            raise UsageException("A new_scopes argument is required.")
        scopes = set() if overwrite else set(self.my_scope_ids())
        if by_id:
            scopes.update(new_scopes)
        else:
            for scope in self.get_scopes_by_name(",".join(new_scopes)):
                scopes.add(scope.get('id'))
            # special case for groups scope since it cannot be resolved by name or inspected by ID via the API
            if LEGACY_GROUPS_SCOPE_NAME in new_scopes:
                scopes.add(LEGACY_GROUPS_SCOPE_ID)
        d = {
            "client": {
                "scopes": list(scopes)
            }
        }

        r = self.client.put('/v2/api/clients/{client_id}'.format(client_id=self.client_id), json_body=d)
        return r.data

    def add_dependent_scopes(self, parent_scope_name, child_scopes,
                             by_id=False, optional=False, requires_refresh_token=False, overwrite=False):
        if not (parent_scope_name and child_scopes):
            raise UsageException("The parent_scope_name and child_scope_names arguments are required.")

        dependent_scopes = []
        new_child_scope_ids = set()
        parent_scopes = self.get_scopes_by_name(parent_scope_name)
        if parent_scopes is None:
            return "no parent scope"
        if len(parent_scopes) != 1:
            return "{sl} parent scopes: {p}".format(sl=str(len(parent_scopes)), p=str(parent_scopes))
        parent_scope_id = parent_scopes[0].get("id")
        if by_id:
            new_child_scope_ids.update(child_scopes)
        else:
            new_child_scopes = self.get_scopes_by_name(",".join(child_scopes))
            for scope in new_child_scopes:
                new_child_scope_ids.add(scope.get('id'))
            # special case for groups scope since it cannot be resolved by name or inspected by ID via the API
            if LEGACY_GROUPS_SCOPE_NAME in child_scopes:
                new_child_scope_ids.add(LEGACY_GROUPS_SCOPE_ID)
        if not overwrite:
            for scope in parent_scopes[0].get('dependent_scopes'):
                if overwrite or scope.get('scope') not in new_child_scope_ids:
                    dependent_scopes.append(scope)
        for scope_id in new_child_scope_ids:
            dependent_scopes.append({'scope': scope_id,
                                     'optional': optional,
                                     'requires_refresh_token': requires_refresh_token})
        d = {
            "scope": {
                "dependent_scopes": dependent_scopes
            }
        }
        print(str(d))
        r = self.client.put('/v2/api/scopes/{i}'.format(i=parent_scope_id),
                            json_body=d)
        return r.data

    def create_scope_with_deps(self, name, description, suffix, dependent_scopes=[], advertised=True,
                               allow_refresh_tokens=True):
        if not (name and description and suffix):
            raise UsageException("The name, description and suffix arguments are required.")
        dependent_scope_arg = []
        if len(dependent_scopes) > 0:
            child_scopes = self.get_scopes_by_name(",".join(dependent_scopes))
            # special case for groups scope since it cannot be resolved by name or inspected by ID via the API
            if LEGACY_GROUPS_SCOPE_NAME in dependent_scopes:
                child_scopes.add({"id": LEGACY_GROUPS_SCOPE_ID})
            for scope in child_scopes:
                dependent_scope_arg.append({
                    "scope": scope.get("id"),
                    "optional": False,
                    "requires_refresh_token": False
                })
        scope = {
            "scope": {
                "name": name,
                "description": description,
                "scope_suffix": suffix,
                "dependent_scopes": dependent_scope_arg,
                "advertised": advertised,
                "allows_refresh_tokens": allow_refresh_tokens
            }
        }

        r = self.client.post("/v2/api/clients/{client_id}/scopes".format(client_id=self.client_id),
                             json_body=scope)
        return r.data

    def delete_scope(self, scope_name):
        if not scope_name:
            raise UsageException("A scope_name argument is required.")
        scopes = self.get_scopes_by_name(scope_name)
        if scopes is None or len(scopes) != 1:
            return "null or multiple scopes"
        scope_id = scopes[0].get('id')
        if scope_id is None:
            return "no scope id"
        r = self.client.delete('/v2/api/scopes/{scope_id}'.format(scope_id=scope_id))
        return r.data

    def get_dependent_scopes(self, scope):
        if not scope:
            raise UsageException("A supported scope argument is required.")
        result = {"scope_string": scope.get("scope_string"), "dependent_scopes": []}
        for ds in scope.get("dependent_scopes"):
            ds_id = ds.get('scope')
            ds_info = {"id": ds_id}
            d = self.get_scopes_by_id(ds_id)
            if d is not None:
                ds_info['scope_string'] = d[0].get('scope_string')
            result['dependent_scopes'].append(ds_info)
        return result


class DerivaJSONTokenStorage(object):
    """
    Stores tokens in json format on disk in the local directory by default.
    """
    def __init__(self, filename=None, permission=None):
        self.filename = filename or DEFAULT_GLOBUS_CREDENTIAL_FILE
        self.permission = permission or stat.S_IRUSR | stat.S_IWUSR

    def write_tokens(self, tokens, overwrite=False):
        all_tokens = self.read_tokens() if not overwrite else dict()
        all_tokens.update(tokens)
        with open(self.filename, 'w+') as fh:
            json.dump(all_tokens, fh, indent=2)

    def read_tokens(self):
        if not os.path.exists(self.filename):
            return dict()
        with open(self.filename) as fh:
            return json.load(fh)

    def clear_tokens(self, requested_scopes=()):
        if not requested_scopes and os.path.exists(self.filename):
            os.remove(self.filename)
            return

        tokens = self.read_tokens() or dict()
        token_set = list()
        for resource, token in tokens.items():
            scope = token["scope"]
            if scope in requested_scopes:
                token_set.append(resource)
        for resource in token_set:
            logging.info("Clearing token for resource: %s" % resource)
            del tokens[resource]
        self.write_tokens(tokens, overwrite=True)


class GlobusNativeLogin:
    def __init__(self, **kwargs):
        self.client = None
        self.client_id = kwargs.get("client_id") or NATIVE_APP_CLIENT_ID
        self.hosts = kwargs.get("hosts")
        self.config_file = kwargs.get("config_file")
        self.exclude_defaults = kwargs.get("exclude_defaults")
        self.default_scopes = DEFAULT_SCOPES.copy()

        try:
            global GLOBUS_SDK, NATIVE_LOGIN
            GLOBUS_SDK = importlib.import_module("globus_sdk")
            NATIVE_LOGIN = importlib.import_module("fair_research_login")
        except Exception as e:
            raise DependencyError("Unable to load a required module: %s" % format_exception(e))

        try:
            storage_file = 'globus-credential%s.json' % \
                           (("-" + self.client_id) if self.client_id != NATIVE_APP_CLIENT_ID else "")
            storage = DerivaJSONTokenStorage(filename=os.path.join(DEFAULT_CONFIG_PATH, storage_file))
            self.client = NATIVE_LOGIN.NativeClient(
                client_id=self.client_id,
                token_storage=storage,
                app_name="Login from deriva-client on %s [%s]%s" %
                         (platform.uname()[1],
                          platform.platform(aliased=True),
                          " to hosts [%s]" % ", ".join(self.hosts) if self.hosts else ""),
                default_scopes=self.default_scopes if not self.exclude_defaults else None)
        except Exception as e:
            logging.error("Unable to instantiate a required class: %s" % format_exception(e))

    def user_info(self, host):
        pass

    def is_logged_in(self, hosts=None, requested_scopes=(), exclude_defaults=False):
        try:
            scopes = set(requested_scopes)
            scopes.update(self.scope_set_from_scope_map(self.hosts_to_scope_map(hosts)))
            if not exclude_defaults:
                scopes.update(self.default_scopes)
            self.client.load_tokens(scopes)
            return True
        except NATIVE_LOGIN.LoadError:
            return False

    def hosts_to_scope_map(self, hosts,
                           match_scope_tag=None,
                           all_tagged_scopes=False,
                           force_refresh=False):
        scope_map = dict()
        for host in hosts:
            scope_map.update({host: []})
            scopes = get_oauth_scopes_for_host(host,
                                               config_file=self.config_file,
                                               force_refresh=force_refresh,
                                               warn_on_discovery_failure=True)
            scope_list = list()
            if scopes:
                for k, v in scopes.items():
                    if match_scope_tag is None:
                        scope_list.append(v)
                        if all_tagged_scopes:
                            continue
                        break
                    elif match_scope_tag == k:
                        scope_list.append(v)
                        break
                    else:
                        continue
            scope_map.update({host: scope_list})

        return scope_map

    @staticmethod
    def scope_set_from_scope_map(scope_map):
        return set([scope for scope_list in [scope_list for scope_list in scope_map.values()] for scope in scope_list])

    @staticmethod
    def host_to_url(host, path="/"):
        if not host:
            return None
        upr = urlparse(host)
        if upr.scheme and upr.netloc:
            url = urljoin(host, path)
        else:
            url = "https://%s%s" % (host, path if not host.endswith("/") else "")
        return url.lower()

    @staticmethod
    def find_access_token_for_host(host, scope_map, tokens, match_scope_tag=None):
        scope = None
        scopes = scope_map.get(host)
        if match_scope_tag:
            scope = scopes[0]
        else:
            for entry in scopes:
                if host.lower() in entry.lower():
                    scope = entry
        for token in tokens.values():
            token_scope = token.get("scope")
            access_token = token.get("access_token")
            if (token_scope == scope) and (access_token is not None):
                return access_token

    def update_bdbag_keychain(self, token=None, host=None, keychain_file=None, allow_redirects=False, delete=False):
        if (token is None) and (host is None):
            return
        keychain_file = keychain_file or bdbkc.DEFAULT_KEYCHAIN_FILE
        entry = {
            "uri": self.host_to_url(host),
            "auth_type": "bearer-token",
            "auth_params": {
                "token": token,
                "allow_redirects_with_token": "True" if allow_redirects else "False"
            }
        }
        bdbkc.update_keychain(entry, keychain_file=keychain_file, delete=delete)

    def login(self,
              hosts=None,
              no_local_server=False,
              no_browser=False,
              requested_scopes=(),
              refresh_tokens=None,
              prefill_named_grant=None,
              additional_params=None,
              force=False,
              match_scope_tag=None,
              exclude_defaults=False,
              update_bdbag_keychain=True,
              bdbag_keychain_file=None):
        scopes = set(requested_scopes)
        scope_map = self.hosts_to_scope_map(hosts, match_scope_tag, force_refresh=True)
        scopes.update(self.scope_set_from_scope_map(scope_map))
        if not exclude_defaults:
            scopes.update(self.default_scopes)
        if not prefill_named_grant:
            prefill_named_grant = self.client.app_name + " with requested scopes [%s] " % ", ".join(scopes)
        tokens = self.client.login(no_local_server=no_local_server,
                                   no_browser=no_browser,
                                   requested_scopes=scopes,
                                   refresh_tokens=refresh_tokens,
                                   prefill_named_grant=prefill_named_grant,
                                   additional_params=additional_params,
                                   force=force)
        if update_bdbag_keychain:
            for host in hosts:
                access_token = self.find_access_token_for_host(host, scope_map, tokens, match_scope_tag=match_scope_tag)
                if access_token:
                    self.update_bdbag_keychain(token=access_token, host=host, keychain_file=bdbag_keychain_file)
        return tokens

    def logout(self, hosts, requested_scopes=(), exclude_defaults=False, bdbag_keychain_file=None):
        tokens = self.client._load_raw_tokens()

        scopes = set(requested_scopes)
        scope_map = self.hosts_to_scope_map(hosts)
        scopes.update(self.scope_set_from_scope_map(scope_map))

        if not exclude_defaults:
            scopes.update(self.default_scopes)
        if not scopes:
            logging.info("Logging out and invalidating tokens for ALL existing scopes.")
            self.client.revoke_token_set(tokens)
            self.client.token_storage.clear_tokens()
            for token in tokens.values():
                self.update_bdbag_keychain(token=token["access_token"], keychain_file=bdbag_keychain_file, delete=True)
            return

        for host in hosts:
            self.update_bdbag_keychain(host=host, keychain_file=bdbag_keychain_file, delete=True)

        token_set = dict()
        for resource, token in tokens.items():
            if token["scope"] in scopes:
                token_set[resource] = token
        self.client.revoke_token_set(token_set)
        self.client.token_storage.clear_tokens(scopes)


class DerivaGlobusAuthUtilCLIException(Exception):
    def __init__(self, message):
        super(DerivaGlobusAuthUtilCLIException, self).__init__(message)


class DerivaGlobusAuthUtilCLI(BaseCLI):
    """Deriva GlobusClientUtil Command-line Interface.
    """

    def __init__(self, *args, **kwargs):
        super(DerivaGlobusAuthUtilCLI, self).__init__(*args, **kwargs)

        self.gau = None
        self.gnl = None
        self.remove_options(['--host', '--credential-file', '--token', '--oauth2-token'])
        self.parser.add_argument("--pretty", "-p", action="store_true",
                                 help="Pretty-print all result output.")
        parent_mutex_group = self.parser.add_mutually_exclusive_group()
        parent_mutex_group.add_argument('--credential-file', '-f', metavar='<file>',
                                        help="Path to a credential file.")
        parent_mutex_group.add_argument('--client-id', '-c', metavar='<client id>',
                                        help="Globus Auth Client ID")
        self.parser.add_argument('--client-secret', '-k', metavar='<client secret key>',
                                 help="Globus Auth Client Secret")

        self.subparsers = self.parser.add_subparsers(title='sub-commands', dest='subcmd')

        # init subparsers and corresponding functions
        self.get_scopes_init()
        self.put_scope_init()
        self.add_scopes_init()
        self.create_scope_init()
        self.delete_scope_init()
        self.get_client_init()
        self.put_client_init()
        self.create_client_init()
        self.delete_client_init()
        self.client_fqdn_init()
        self.token_init()
        self.login_init()
        self.logout_init()
        # self.user_info_init()

    def get_scopes_init(self):
        def get_scopes(args):
            if args.scope_ids:
                return self.gau.get_scopes_by_id(args.scope_ids)
            elif args.scope_names:
                return self.gau.get_scopes_by_name(args.scope_names)
            else:
                return self.gau.list_all_scopes()

        parser = self.subparsers.add_parser(
            'get-scopes',
            help="Get one or more scopes by name or ID, or all scopes owned by this client if no scope ID is "
                 "specified.")
        parser.add_argument("--scope-ids", metavar="[scopes]",
                            help="A comma-delimited list of scope IDs to retrieve.")
        parser.add_argument("--scope-names", metavar="[scopes]",
                            help="A comma-delimited list of scope names to retrieve.")
        parser.set_defaults(func=get_scopes)

    def put_scope_init(self):
        def put_scope(args):
            if args.scope_id:
                return self.gau.update_scope(args.scope_id, args.scope_config)
            else:
                return self.gau.create_scope(args.scope_config)

        parser = self.subparsers.add_parser(
            "put-scope",
            help="Create or update a scope owned by this client from a valid JSON scope configuration in string form, "
                 "or a path to an equivalent file.")
        parser.add_argument("--scope-id", metavar="<scope id>", help="The scope ID.")
        parser.add_argument("scope_config", metavar="<JSON String or path to file>",
                            help="A valid JSON scope configuration in string form, or a path to an equivalent file.")
        parser.set_defaults(func=put_scope)

    def add_scopes_init(self):
        def add_scopes(args):
            if args.parent_scope:
                return self.gau.add_dependent_scopes(args.parent_scope,
                                                     args.scope_names,
                                                     args.by_id,
                                                     args.optional,
                                                     args.requires_refresh_token,
                                                     args.overwrite)
            else:
                return self.gau.add_scopes(args.scope_names, args.by_id)

        parser = self.subparsers.add_parser(
            'add-scopes',
            help="Add one or more scopes by name (or optionally by ID) to this client, or add one or more scopes as "
                 "dependent scopes of a specified parent scope.")
        parser.add_argument("scope_names", metavar="[scopes]", default=list(),
                            type=lambda s: [item.strip() for item in s.split(',')],
                            help="A comma-delimited list of scope names or scope IDs.")
        parser.add_argument("--parent-scope", metavar="<scope name>",
                            help="The parent scope name if adding dependent scopes.")
        parser.add_argument("--by-id", action="store_true",
                            help="Interpret the scope name list as a list of scope IDs. "
                                 "Note: does not validate the listed scopes beforehand. Default false.")
        parser.add_argument("--optional", action="store_true",
                            help="When adding to dependent scopes, sets the scope as optional. Default false.")
        parser.add_argument("--requires-refresh-token", action="store_true",
                            help="When adding to dependent scopes, sets the scope to require the usage of refresh "
                                 "tokens. Default false.")
        parser.add_argument("--overwrite", action="store_true",
                            help="Overwrite any existing scopes. Default false.")
        parser.set_defaults(func=add_scopes)

    def create_scope_init(self):
        def create_scope(args):
            return self.gau.create_scope_with_deps(args.name,
                                                   args.description,
                                                   args.suffix,
                                                   args.dependent_scope_names,
                                                   args.advertised,
                                                   args.allow_refresh_token)
        parser = self.subparsers.add_parser(
            'create-scope',
            help="Creates a scope for each registered FQDN + the id of the client from command-line parameters.")
        parser.add_argument("name", metavar="<scope name>",
                            help="The new scope name. Max 100 chars.")
        parser.add_argument("description", metavar="<scope desc>",
                            help="A scope description. Max 5000 chars.")
        parser.add_argument("suffix", metavar="<scope suffix>",
                            help="String consisting of lowercase letters, number, and underscores.")
        parser.add_argument("--dependent-scope-names", metavar="[scopes]", default=list(),
                            type=lambda s: [item.strip() for item in s.split(',')],
                            help="A comma-delimited list of dependent scope names.")
        parser.add_argument("--advertised", action="store_true",
                            help="Whether or not the scope should show up in searches. Default: True")
        parser.add_argument("--allow-refresh-token", action="store_true",
                            help="Whether or not the scope allows refresh tokens to be issued. Default: True")
        parser.set_defaults(func=create_scope)

    def delete_scope_init(self):
        def delete_scope(args):
            return self.gau.delete_scope(args.scope_name)

        parser = self.subparsers.add_parser(
            "del-scope",
            help="Delete the specified scope name. Deleting a scope deletes all resources associated with it. This "
                 "operation can cause other apps and services that depend on the scope to stop working. This action "
                 "cannot be undone.")
        parser.add_argument("scope_name", metavar="<scope_name>", help="The scope name to delete.")
        parser.set_defaults(func=delete_scope)

    def get_client_init(self):
        def get_client(args):
            if args.get_client_id:
                return self.gau.get_client(args.get_client_id)
            else:
                return self.gau.get_clients()

        parser = self.subparsers.add_parser(
            'get-client',
            help="Retrieve client information for the specified client, or all clients owned by this client ID if no "
                 "ID specified.")
        parser.add_argument("--get-client-id", metavar="<client ID>",
                            help="Retrieve client information for the specified client ID.")
        parser.set_defaults(func=get_client)

    def put_client_init(self):
        def put_client(args):
            if args.create:
                return self.gau.new_client(args.client_config)
            else:
                return self.gau.update_client(args.client_config, args.put_client_id)

        parser = self.subparsers.add_parser(
            'put-client',
            help="Update this (or another) client OR create a child client of this client from a valid JSON client "
                 "configuration in string form, or a path to an equivalent file.")
        parser.add_argument("client_config", metavar="<JSON String or path to file>",
                            help="A valid JSON client config in string form, or a path to an equivalent file.")
        parser.add_argument("--put-client-id", metavar="<client id>",
                            help="The client ID to update, or implicitly this client's ID if not specified.")
        parser.add_argument("--create", action="store_true",
                            help="Create a new child client from the input client config.")
        parser.set_defaults(func=put_client)

    def create_client_init(self):
        def create_client(args):
            return self.gau.create_client(args.name,
                                          args.redirect_uris,
                                          args.public,
                                          args.visibility,
                                          args.project,
                                          args.required_idp,
                                          args.preselect_idp,
                                          args.terms_of_service,
                                          args.privacy_policy)

        parser = self.subparsers.add_parser('create-client', help="Create a client from command-line parameters.")
        parser.add_argument("name", metavar="<client name>",
                            help="Display name shown to users in consents. String without line-breaks, with no "
                                 "more than 100 characters.")
        parser.add_argument("redirect_uris", metavar="[redirect uris]", default=list(),
                            type=lambda s: [item.strip() for item in s.split(',')],
                            help="A comma-delimited list of URIs that may be used in OAuth authorization flows.")
        parser.add_argument("--public", action="store_true",
                            help="Create a public (native app) client.")
        parser.add_argument("--visibility", choices=["public", "private"],
                            help="\"private\" means that only entities in the same project as the client "
                                 "can view it. \"public\" means that any authenticated entity can view it.")
        parser.add_argument("--project", metavar="<project ID>",
                            help="ID representing the project this client belongs to.")
        parser.add_argument("--required-idp", metavar="<IDP ID>",
                            help="ID representing an Identity Provider. In order to use this client a user must"
                                 " have an identity from this IdP in their identity set.")
        parser.add_argument("--preselect-idp", metavar="<IDP ID>",
                            help="ID representing an Identity Provider. This preselects the given IdP on the "
                                 "Globus Auth login page if the user is not already authenticated.")
        parser.add_argument("--terms-of-service", metavar="<URL>",
                            help="A URL to the terms and conditions statement for this client.")
        parser.add_argument("--privacy_policy", metavar="<URL>",
                            help="A URL to the privacy policy for this client.")
        parser.set_defaults(func=create_client)

    def delete_client_init(self):
        def delete_client(args):
            return self.gau.delete_client(args.del_client_id)

        parser = self.subparsers.add_parser(
            'del-client',
            help="Delete a client by ID. Warning: deletes all resources associated with it the client. This includes "
                 "user consents, scopes (which means this operation can cause other apps and services that depend "
                 "on those scopes to stop working as well), and any child clients owned by the client (which in turn "
                 "means that all resources associated with the child clients would get deleted as well). This action "
                 "cannot be undone.")
        parser.add_argument("del_client_id", metavar="<client id>",
                            help="The client ID to update, or implicitly this client's ID if not specified.")
        parser.set_defaults(func=delete_client)

    def client_fqdn_init(self):
        def client_fqdn(args):
            if args.add:
                return self.gau.add_fqdn_to_client(args.fqdn)
            else:
                return self.gau.get_client_for_fqdn(args.fqdn)

        parser = self.subparsers.add_parser(
            'client-fqdn',
            help="Retrieve client information for an FQDN, or add and FQDN to this client.")
        parser.add_argument("--add", action="store_true",
                            help="Add the specified FQDN to this client ID")
        parser.add_argument("fqdn", metavar="<fqdn>",
                            help="The fully qualified domain name to lookup or add to this client.")
        parser.set_defaults(func=client_fqdn)

    def token_init(self):
        def token(args):
            if args.validate:
                return self.gau.verify_access_token(args.token)
            elif args.dependent:
                return self.gau.get_dependent_access_tokens(args.token, args.refresh)
            else:
                return self.gau.introspect_access_token(args.token)

        parser = self.subparsers.add_parser('token', help="Introspect or validate an access token.")
        parser.add_argument("--validate", action="store_true",
                            help="Validate the access token.")
        parser.add_argument("--dependent", action="store_true",
                            help="Get dependent access token(s).")
        parser.add_argument("--refresh", action="store_true",
                            help="Request refresh tokens when getting dependent access token(s).")
        parser.add_argument("token", metavar="<token>",
                            help="The access token to introspect (or validate).")
        parser.set_defaults(func=token)

    def login_init(self):
        def login(args):
            if self.gnl.is_logged_in(args.hosts, args.requested_scopes,
                                     exclude_defaults=args.exclude_defaults) and not args.force:
                return "You are already logged in."
            else:
                response = self.gnl.login(hosts=args.hosts,
                                          no_local_server=args.no_local_server or args.no_browser,
                                          no_browser=args.no_browser,
                                          refresh_tokens=args.refresh,
                                          force=args.force,
                                          requested_scopes=args.requested_scopes,
                                          match_scope_tag=args.match_scope_tag,
                                          exclude_defaults=args.exclude_defaults,
                                          update_bdbag_keychain=not args.no_bdbag_keychain,
                                          bdbag_keychain_file=args.bdbag_keychain_file)
                if args.show_tokens:
                    return response
                else:
                    return "Login Successful"

        parser = self.subparsers.add_parser('login', help="Login with Globus Auth")
        mutex_group = parser.add_mutually_exclusive_group(required=True)
        mutex_group.add_argument("--hosts", metavar="[hostnames]", default=list(),
                                 type=lambda s: [item.strip() for item in s.split(',')],
                                 help="A comma-delimited list of host names to login to. "
                                 "An attempt to determine the required scope will be made by checking the local "
                                 "configuration or (if required) contacting each <host> will be made.")
        mutex_group.add_argument("--requested-scopes", metavar="[scopes]", default=list(),
                                 type=lambda s: [item.strip() for item in s.split(',')],
                                 help="A comma-delimited list of scope names to request tokens for. "
                                 "If not specified, an attempt will be made to determine the required scope by "
                                 "checking the local configuration or (if required) contacting each <host>.")

        parser.add_argument('--client-id', '-c', metavar='<client id>',
                            help="Use a different client ID than the default.")
        parser.add_argument("--no-local-server", action="store_true",
                            help="Do not launch a local server to receive the authorization redirect response. You "
                                 "will be required to input the response token from the authorization request.")
        parser.add_argument("--no-browser", action="store_true",
                            help="Do not launch a browser instance on this system for initiating the login flow. "
                                 "Use this argument on a \"headless\" system without a GUI environment where a web "
                                 "browser is not available.")
        parser.add_argument("--no-bdbag-keychain", action="store_true",
                            help="Do not update the bdbag keychain file with result access tokens. Default false.")
        parser.add_argument('--bdbag-keychain-file', metavar='<file>',
                            help="Non-default path to a bdbag keychain file.")
        parser.add_argument("--refresh", action="store_true",
                            help="Enable the use of refresh tokens to extend the login time until revoked.")
        parser.add_argument("--force", action="store_true",
                            help="Force a login flow even if the current access token set is valid.")
        parser.add_argument("--show-tokens", action="store_true",
                            help="Display the tokens from the authorization response.")
        parser.add_argument("--match-scope-tag", metavar="<scope tag>",
                            help="When using the --hosts argument, specify an explicit scope tag to match against the "
                                 "list of configured scopes for the given host(s).")
        parser.add_argument("--exclude-defaults", action="store_true",
                            help="In addition to any specified scopes or host-to-scope mappings, do not include the "
                                 "default set of scopes: [%s]" % ", ".join(DEFAULT_SCOPES))
        parser.set_defaults(func=login)

    def logout_init(self):
        def logout(args):
            self.gnl.logout(args.hosts,
                            args.requested_scopes,
                            args.exclude_defaults,
                            args.bdbag_keychain_file)
            return "You have been logged out."

        parser = self.subparsers.add_parser("logout", help="Revoke and clear tokens. If no arguments are specified, "
                                                           "all tokens will be removed.")
        mutex_group = parser.add_mutually_exclusive_group()
        mutex_group.add_argument("--hosts", metavar="[hostnames]", default=list(),
                                 type=lambda s: [item.strip() for item in s.split(',')],
                                 help="A comma-delimited list of host names to revoke tokens for. "
                                 "An attempt to determine the associated scope(s) will be made by checking the local "
                                 "configuration or (if required) contacting each <host> will be made.")
        mutex_group.add_argument("--requested-scopes", metavar="[scopes]", default=list(),
                                 type=lambda s: [item.strip() for item in s.split(',')],
                                 help="A comma-delimited list of scope names to revoke tokens for. "
                                 "If not specified, an attempt will be made to determine the associated scope(s) by "
                                 "checking the local configuration or (if required) contacting each <host>.")
        parser.add_argument("--exclude-defaults", action="store_true",
                            help="In addition to any specified scopes or host-to-scope mappings, do not include the "
                                 "default set of scopes: [%s]" % ", ".join(DEFAULT_SCOPES))
        parser.add_argument('--bdbag-keychain-file', metavar='<file>',
                            help="Non-default path to a bdbag keychain file.")
        parser.set_defaults(func=logout)

    def user_info_init(self):
        def user_info(args):
            if self.gnl.is_logged_in(args.hosts, args.requested_scopes):
                return self.gnl.user_info()
            else:
                return "Login required."

        parser = self.subparsers.add_parser("user-info", help="Display information for the currently logged-in user.")
        parser.set_defaults(func=user_info)

    def main(self):
        args = self.parse_cli()

        def _cmd_error_message(emsg):
            return "{prog} {subcmd}: {msg}".format(
                prog=self.parser.prog, subcmd=args.subcmd, msg=emsg)

        try:
            if not hasattr(args, 'func'):
                self.parser.print_usage()
                return 2

            if args.subcmd == "login" or args.subcmd == "logout" or args.subcmd == "user-info":
                self.gnl = GlobusNativeLogin(**vars(args))
            else:
                self.gau = GlobusAuthUtil(**vars(args))

            response = args.func(args)
            if args.pretty:
                if isinstance(response, dict) or isinstance(response, list):
                    try:
                        print(json.dumps(response, indent=2))
                        return
                    except:
                        pprint(response)
                        return
                elif not isinstance(response, str):
                    pprint(response)
                    return
            print(response)
            return

        except UsageException as e:
            eprint("{prog} {subcmd}: {msg}".format(prog=self.parser.prog, subcmd=args.subcmd, msg=e))
        except ConnectionError as e:
            eprint("{prog}: Connection error occurred".format(prog=self.parser.prog))
        except GlobusAuthAPIError as e:
            if 401 == e.http_status:
                msg = 'Authentication required: %s' % e.message
            elif 403 == e.http_status:
                msg = 'Permission denied: %s' % e.message
            else:
                msg = e
            logging.debug(format_exception(e))
            eprint(_cmd_error_message(msg))
        except DependencyError as e:
            eprint(_cmd_error_message(e))
        except RuntimeError as e:
            logging.debug(format_exception(e))
            eprint('An unexpected runtime error occurred')
        except:
            eprint('An unexpected error occurred')
            traceback.print_exc()
        return 1


def main():
    desc = "DERIVA Globus Auth Utilities"
    info = "For more information see: https://github.com/informatics-isi-edu/deriva-py"
    return DerivaGlobusAuthUtilCLI(desc, info).main()


if __name__ == '__main__':
    sys.exit(main())



import os
import logging
import uuid
from datetime import datetime
from importlib import import_module
from deriva.core import get_credential, urlsplit, urlunsplit, format_exception, stob
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.base_processor import *


class UploadPostProcessor(BaseProcessor):
    """
    Post processor that transfers download results to remote systems.
    """

    def __init__(self, envars=None, **kwargs):
        super(UploadPostProcessor, self).__init__(envars, **kwargs)
        self.scheme = None
        self.netloc = None
        self.path = None
        self.credentials = None

    def process(self):
        target_url_param = "target_url"
        target_url = self.parameters.get(target_url_param)
        if not target_url:
            raise DerivaDownloadConfigurationError(
                "%s is missing required parameter '%s' from %s" %
                (self.__class__.__name__, target_url_param, PROCESSOR_PARAMS_KEY))
        if self.envars:
            target_url = target_url.format(**self.envars)
        target_url = target_url.strip(" ")
        upr = urlsplit(target_url, "https")
        self.scheme = upr.scheme.lower()
        self.netloc = upr.netloc
        self.path = upr.path.strip("/")
        host = urlunsplit((self.scheme, upr.netloc, "", "", ""))
        creds = get_credential(host)
        if not creds:
            logging.info("Unable to locate credential entry for: %s" % host)
        self.credentials = creds or dict()

        return self.outputs


class Boto3UploadPostProcessor(UploadPostProcessor):
    BOTO3 = None
    BOTOCORE = None

    def import_boto3(self):
        # locate library
        if self.BOTO3 is None and self.BOTOCORE is None:
            try:
                self.BOTO3 = import_module("boto3")
                self.BOTOCORE = import_module("botocore")
            except ImportError as e:
                raise DerivaDownloadConfigurationError("Unable to find required module. "
                                                       "Ensure that the Python package \"boto3\" is installed.", e)
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            self.BOTO3.set_stream_logger('')

    def __init__(self, envars=None, **kwargs):
        super(Boto3UploadPostProcessor, self).__init__(envars, **kwargs)
        self.import_boto3()

    def process(self):
        super(Boto3UploadPostProcessor, self).process()
        key = self.credentials.get("key")
        secret = self.credentials.get("secret")
        token = self.credentials.get("token")
        role_arn = self.parameters.get("role_arn")
        profile_name = self.parameters.get("profile")
        region_name = self.parameters.get("region")
        try:
            session = self.BOTO3.session.Session(profile_name=profile_name, region_name=region_name)
        except Exception as e:
            raise DerivaDownloadConfigurationError("Unable to create Boto3 session: %s" % format_exception(e))

        if role_arn:
            try:
                sts = session.client('sts')
                response = sts.assume_role(RoleArn=role_arn, RoleSessionName='DERIVA-Export', DurationSeconds=3600)
                temp_credentials = response['Credentials']
                key = temp_credentials['AccessKeyId']
                secret = temp_credentials['SecretAccessKey']
                token = temp_credentials['SessionToken']
            except Exception as e:
                raise RuntimeError("Unable to get temporary credentials using arn [%s]. %s" %
                                   (role_arn, get_typed_exception(e)))

        try:
            if self.scheme == "gs":
                endpoint_url = "https://storage.googleapis.com"
                config = self.BOTO3.session.Config(signature_version="s3v4")
                kwargs = {"aws_access_key_id": key,
                          "aws_secret_access_key": secret,
                          "endpoint_url": endpoint_url,
                          "config": config}
            else:
                kwargs = {"aws_access_key_id": key, "aws_secret_access_key": secret}
                if token:
                    kwargs.update({"aws_session_token": token})

            s3_client = session.client("s3", **kwargs)
            kwargs["config"] = self.BOTO3.session.Config(signature_version=self.BOTOCORE.UNSIGNED)
            s3_client_unsigned = self.BOTO3.client('s3', **kwargs)
        except Exception as e:
            raise DerivaDownloadError("Unable to create Boto3 storage client: %s" % format_exception(e))

        bucket_name = self.netloc
        bucket_exists = True
        try:
            s3_client.head_bucket(Bucket=bucket_name)
        except self.BOTOCORE.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                bucket_exists = False
        except Exception as e:
            raise DerivaDownloadError("Unable to query target bucket: %s" % format_exception(e))

        if not bucket_exists:
            raise DerivaDownloadError("Target bucket [%s] does not exist." % bucket_name)

        object_qualifier = os.path.basename(self.identity.get("id", "")) or "anon-" + str(uuid.uuid4())
        if not stob(self.parameters.get("overwrite", "False")):
            object_qualifier = "/".join([object_qualifier, datetime.strftime(datetime.now(), "%Y-%m-%d_%H.%M.%S")])

        for k, v in self.outputs.items():
            object_name = "/".join([self.path, object_qualifier, k])
            file_path = v[LOCAL_PATH_KEY]
            acl = self.parameters.get("acl", "private")
            signed_url = stob(self.parameters.get("signed_url", acl == "public-read"))
            if signed_url:
                client = s3_client_unsigned if acl == "public-read" else s3_client
                remote_path = client.generate_presigned_url(
                    'get_object', Params={'Bucket': bucket_name, 'Key': object_name})
            else:
                remote_path = urlunsplit((self.scheme, self.netloc, object_name, "", ""))
            logging.info("Uploading file [%s] to: %s" % (file_path, remote_path))
            remote_paths = v.get(REMOTE_PATHS_KEY, list())
            remote_paths.append(remote_path)
            v[REMOTE_PATHS_KEY] = remote_paths
            self.make_file_output_values(file_path, v)
            with open(file_path, "rb") as input_file:
                try:
                    response = s3_client.put_object(ACL=acl,
                                                    Bucket=bucket_name,
                                                    Key=object_name,
                                                    Body=input_file,
                                                    ContentType=v[CONTENT_TYPE_KEY],
                                                    ContentLength=v[FILE_SIZE_KEY],
                                                    ContentMD5=v[MD5_KEY][1],
                                                    Metadata={"Content-MD5": v[MD5_KEY][0]})
                except Exception as e:
                    raise DerivaDownloadError("Upload of %s failed: %s" % (remote_path, format_exception(e)))

        return self.outputs



"""Definitions and implementation for validating ERMrest schema annotations."""

import json
import logging
import pkgutil
import jsonschema
import warnings

from .. import core
from . import tag

logger = logging.getLogger(__name__)


class _AnnotationSchemas (object):
    """Container of annotation schemas."""

    # TODO: inherit from 'collections.abc.Mapping'

    def __init__(self):
        super(_AnnotationSchemas, self).__init__()
        self._abbrev = dict((v, k) for k, v in tag.items())
        self._schemas = {}

    def __getitem__(self, key):
        try:
            return self._schemas[key]
        except KeyError:
            abbrev = self._abbrev[key]  # let this raise a 'KeyError' if the tag name is unknown
            s = pkgutil.get_data(core.__name__, 'schemas/%s.schema.json' % abbrev).decode()
            v = json.loads(s)
            self._schemas[key] = v
            return v

    def __iter__(self):
        return iter(self._schemas)

    def __len__(self):
        return len(self._schemas)


_schemas = _AnnotationSchemas()
_schema_store = {}
try:
    _schema_store = {  # the schema store for the schema resolver, stores all schemas that are extended
        _schemas[tag_name]['$id']: _schemas[tag_name] for tag_name in [tag.export, tag.source_definitions]
    }
except FileNotFoundError as e:
    warnings.warn("Unable to read base schemas. Error: %s" % e)


def _nop(validator, value, instance, schema):
    """NOP validator function."""
    return


def validate(model_obj, tag_name=None, validate_model_names=True):
    """Validate the annotation(s) of the model object.

    :param model_obj: model object container of annotations
    :param tag_name: tag name of the annotation to validate, if none, will validate all known annotations
    :param validate_model_names: validate model names used in annotations
    :return: a list of validation errors, if any
    """
    assert hasattr(model_obj, 'annotations')
    if not tag_name:
        errors = []
        for tag_name in model_obj.annotations:
            errors.extend(_validate(model_obj, tag_name, validate_model_names))
        return errors
    elif tag_name in model_obj.annotations:
        return _validate(model_obj, tag_name, validate_model_names)
    else:
        return []


def _validate(model_obj, tag_name, validate_model_names):
    """Validate an annotation of the model object.

    :param model_obj: model object container of annotations
    :param tag_name: tag name of the annotation to validate
    :param validate_model_names: validate model names used in annotations
    :return: a list of validation errors, if any
    """
    assert tag_name in model_obj.annotations
    logger.debug("Validating '%s' against schema for '%s'" % (_printable_name(model_obj), tag_name))
    try:
        schema = _schemas[tag_name]
        resolver = jsonschema.RefResolver.from_schema(schema, store=_schema_store)
        if validate_model_names:
            ExtendedValidator = jsonschema.validators.extend(
                jsonschema.Draft7Validator,
                {
                    'valid-table': _validate_table_fn(model_obj),
                    'valid-column': _validate_column_fn(model_obj),
                    'valid-constraint': _validate_constraint_fn(model_obj),
                    'valid-source-key': _validate_source_key_fn(model_obj),
                    'valid-source-path': _validate_source_path_fn(model_obj)
                }
            )
            validator = ExtendedValidator(schema, resolver=resolver)
            validator.validate(model_obj.annotations[tag_name])
        else:
            jsonschema.validate(model_obj.annotations[tag_name], schema, resolver=resolver)
    except jsonschema.ValidationError as e:
        logger.error("Failed to validate '%s' for annotation '%s'" % (_printable_name(model_obj), tag_name))
        logger.error(e)
        return [e]
    except KeyError as e:
        logger.error("Failed to validate '%s' for annotation '%s'" % (_printable_name(model_obj), tag_name))
        msg = 'Unknown annotation tag name "%s"' % tag_name
        logger.error(msg)
        return [jsonschema.ValidationError(msg, cause=e)]
    except FileNotFoundError as e:
        logger.warning('No schema document found for tag name %s : %s' % (tag_name, e))
    return []


def _printable_name(model_obj):
    """Returns a print-frendly name for a model object.

    :param model_obj: a model, schema, table, column, key, or foreign key object.
    :return: string representation of its name or "catalog" if no name found
    """
    if not hasattr(model_obj, 'name'):
        return 'catalog'
    if hasattr(model_obj, 'constraint_name'):
        return "%s:%s" % model_obj.name
    if hasattr(model_obj, 'table'):
        return "%s:%s:%s" % (model_obj.table.schema.name, model_obj.table.name, model_obj.name)
    if hasattr(model_obj, 'schema'):
        return "%s:%s" % (model_obj.schema.name, model_obj.name)
    if hasattr(model_obj, 'name'):
        return model_obj.name
    return 'unnamed model object'


def _is_qualified_name(value):
    """Tests if the given value looks like a schema-qualified name."""
    return isinstance(value, list) and len(value) == 2 and all(isinstance(item, str) for item in value)


def _validate_column_fn(model_obj):
    """Produces a column name validation function for the model object

    :param model_obj: expects column object
    :return: validation function
    """
    if hasattr(model_obj, 'table'):
        model_obj = model_obj.table
    if not hasattr(model_obj, 'column_definitions'):
        return _nop

    def _validation_func(validator, value, instance, schema):
        if not (value and isinstance(instance, str)):
            return

        try:
            model_obj.column_definitions[instance]
        except KeyError as e:
            raise jsonschema.ValidationError("'%s' not found in column definitions" % instance, cause=e,
                                             validator=validator, validator_value=value,
                                             instance=instance, schema=schema)

    return _validation_func


def _validate_table_fn(model_obj):
    """Produces a table name validation function for the model object

    :param model_obj: expects table object
    :return: validation function
    """
    if not hasattr(model_obj, 'schema'):
        return _nop

    def _validation_func(validator, value, instance, schema):
        if not (value and _is_qualified_name(instance)):
            return

        try:
            model_obj.schema.model.schemas[instance[0]].tables[instance[1]]
        except KeyError as e:
            raise jsonschema.ValidationError("Table %s not found in model" % instance, cause=e,
                                             validator=validator, validator_value=value,
                                             instance=instance, schema=schema)

    return _validation_func


def _validate_constraint_fn(model_obj):
    """Produces a constraint validation function for the model object.

    The directive takes a list value that may include one or all of the following strings:
     - 'inbound': inbound fkey relationships are valid,
     - 'outbound': outbound fkey relationships are valid
     - 'key': keys of the object are valid

    :param model_obj: table object
    :return: validation function
    """
    if not hasattr(model_obj, 'schema'):
        return _nop

    table = model_obj
    model = table.schema.model

    # define a validation function for the model object
    def _validation_func(validator, value, instance, schema):
        if not value or not _is_qualified_name(instance):
            return

        # validate if key
        if 'key' in value:
            try:
                schemaobj = model.schemas[instance[0]]
                key = table.keys[(schemaobj, instance[1])]
                return  # return immediately on validation condition
            except KeyError as e:
                # when no other validation options exist, terminate and raise exception
                if len(value) == 1:
                    raise jsonschema.ValidationError("%s not found in keys of '%s'" % (instance, table.name), cause=e,
                                                     validator=validator, validator_value=value,
                                                     instance=instance, schema=schema)

        # if not already validated as a key, then validate as inbound or outbound fkey
        try:
            if not (
                ('outbound' in value and model.fkey(instance).table == table) or
                ('inbound' in value and model.fkey(instance).pk_table == table)
            ):
                raise jsonschema.ValidationError("%s not related to '%s'" % (instance, table.name),
                                                 validator=validator, validator_value=value,
                                                 instance=instance, schema=schema)
            return
        except KeyError as e:
            raise jsonschema.ValidationError("unable to validate constraint %s" % instance, cause=e,
                                             validator=validator, validator_value=value,
                                             instance=instance, schema=schema)

    return _validation_func


def _validate_source_key_fn(model_obj):
    """Produces a source key validation function for the model object.

    :param model_obj: model object
    :return: validation function
    """
    sourcekeys = model_obj.annotations.get(tag.source_definitions, {}).get('sources', {})

    # define a validation function for the model object
    def _validation_func(validator, value, instance, schema):
        if value and isinstance(instance, str) and instance not in sourcekeys:
            raise jsonschema.ValidationError("'%s' not found in source definitions" % instance,
                                             validator=validator, validator_value=value, instance=instance, schema=schema)

    return _validation_func


def _validate_source_path_fn(model_obj):
    """Produces a source path validation function for the model object.

    :param model_obj: model object
    :return: validation function
    """
    if not (hasattr(model_obj, 'column_definitions') and hasattr(model_obj, 'schema')):
        return _nop

    _model = model_obj.schema.model

    # define a validation function for the model object
    def _validation_func(validator, value, instance, schema):
        if not value:  # 'true' to indicate desire to validate model
            return

        # validate the fkey path case
        if isinstance(instance, list):
            current_table = model_obj  # start the "current" table in the fkey path
            # iterate over the instance of the fkey path
            for item in instance:
                # validate that a column name belongs to the current table
                if isinstance(item, str):
                    if item not in {c.name for c in current_table.column_definitions}:
                        raise jsonschema.ValidationError(
                            "'%s' not found in column definitions of table %s" % (item, [current_table.schema.name, current_table.name]),
                            validator=validator, validator_value=value, instance=instance, schema=schema)
                    else:
                        # source-path should terminate on a column-name, syntactically
                        break

                # validate an inbound or outbound fkey in the path
                elif isinstance(item, dict) and any(_is_qualified_name(item[io]) for io in ['inbound', 'outbound'] if io in item):
                    for direction, match_with_table, update_current_table in [
                        ('outbound',    'table',    'pk_table'),
                        ('inbound',     'pk_table', 'table')
                    ]:
                        if direction in item:
                            constraint_name = item[direction]
                            try:
                                fkey = _model.fkey(constraint_name)  # test if fkey exists in model
                                if getattr(fkey, match_with_table) != current_table:  # test if current table matches fkey table property
                                    raise jsonschema.ValidationError(
                                        "%s foreign key %s not associated with %s" % (direction, constraint_name, [current_table.schema.name, current_table.name]),
                                        validator=validator, validator_value=value, instance=instance, schema=schema)
                                current_table = getattr(fkey, update_current_table)  # update the "current" table
                            except KeyError as e:
                                # fkey not found in model
                                raise jsonschema.ValidationError("%s not found in model fkeys" % constraint_name, cause=e,
                                                                 validator=validator, validator_value=value,
                                                                 instance=instance, schema=schema)
                else:
                    break  # unrecognized source-path syntax, break out and let schema validation take its course
        # and ignore anything unexpected

    return _validation_func



import re
from deriva.core import BaseCLI
import platform

class ConfigUtil:
    @classmethod
    def find_toplevel_node(cls, catalog_config, schema_name, table_name):
        if table_name is not None and schema_name is None:
            raise ValueError("table specified without a schema")

        if schema_name is None:
            return catalog_config

        schema = catalog_config.schemas.get(schema_name)
        if schema is None:
            raise (ValueError("no schema named {s} found.".format(s=schema_name)))

        if table_name is None:
            return schema
        else:
            table = schema.tables.get(table_name)
            if table is None:
                raise (ValueError("no table named {t} in schema {s}".format(s=schema_name, t=table_name)))
            return table

class BaseSpec(dict):
    ATTRIBUTE_TYPES = ["schema", "table", "column", "foreign_key"]

    def __init__(self, specdict, speclist, spectype, validate_node_types_only=False):
        self.specdict = specdict
        self.speclist = speclist
        self.spectype = spectype
        self.node_types = ["schema", "table", "column", "foreign_key", "foreign_key_schema"]
        self.set_keys()        
        if not validate_node_types_only:
            self.validate()
        # self.validate_node_types() # FIXME: this causes an error for catalog annotations which do not have any of attributes listed in self.node_types

    def get_pattern(self, base):
        return self.get(self.make_pat_name(base))

    def is_exact_spec(self):
        for n in self.node_types:
            if self.get_pattern(n) is not None:
                return False
        return True

    def validate(self):
        for k in self.specdict.keys():
            valid = False
            for n in self.node_types:
                if k == n or k == self.make_pat_name(n):
                    valid = True
            if k in self.speclist:
                valid = True
            if not valid:
                raise ValueError("unknown key '{key}' in {t} spec".format(key=k, t=self.spectype))
        return True

    def has_node_key(self, key):
        return key in self.specdict or self.make_pat_name(key) in self.specdict

    def is_node_spec(self):
        for key in self.node_types:
            if key in self.specdict or self.make_pat_name(key) in self.specdict:
                return True
        return False

    def validate_node_types(self):
        if not self.is_node_spec():
            raise ValueError("Unknown node type {s}".format(s=str(self.specdict)))

        if not self.has_node_key('schema'):
            raise ValueError("No schema entry or pattern in spec {s}".format(s=str(self.specdict)))
        if not self.has_node_key('table'):
            if self.has_node_key('column') or self.has_node_key('foreign_key'):
                raise ValueError("No table entry or pattern in spec {s}".format(s=str(self.specdict)))

            if self.has_node_key('foreign_key'):
                if not self.has_node_key('foreign_key_schema'):
                    raise ValueError("foreign_key spec missing foreign_key_schema: {s}".format(s=str(self.specdict)))

    def config_format(self):
        new = dict()
        for k in self.specdict.keys():
            if isinstance(self.specdict[k], re._pattern_type):
                new[k] = self.specdict[k].pattern
            else:
                new[k] = self.specdict[k]
        return new

    def set_keys(self):
        for k in self.specdict.keys():
            self[k] = self.specdict[k]
        for n in self.node_types:
            if self.get_pattern(n) is not None:
                if self.get(n) is not None:
                    raise ValueError(
                        "can't have both '{n}' and '{p}' in the same spec".format(n=n, p=self.make_pat_name(n)))
                self[self.make_pat_name(n)] = re.compile(self.get_pattern(n))

    @classmethod
    def make_pat_name(cls, base):
        return base + '_pattern'

    def matches(self, base, value, exact=False, key=None):
        if key is not None and not self.get(u'uri') == key:
            return False
        if self.get(base) is not None and value == self.get(base):
            return True
        if exact:
            return False
        else:
            pat = self.get_pattern(base)
            return pat is not None and pat.match(value) is not None

    def schema_entry_matches(self, schema_name, exact=False, key=None):
        return self.matches('schema', schema_name, exact=exact, key=key)

    def foreign_key_entry_matches(self, table_schema_name, table_name, fkey_schema_name, fkey_name, exact=False,
                                  key=None):
        return self.table_entry_matches(table_schema_name, table_name, exact=exact, key=key) and self.matches(
            "foreign_key_schema", fkey_schema_name, exact=exact, key=key) and self.matches("foreign_key", fkey_name,
                                                                                           exact=exact, key=key)

    def table_entry_matches(self, schema_name, table_name, exact=False, key=None):
        return self.schema_entry_matches(schema_name, exact=exact, key=key) and self.matches("table", table_name,
                                                                                             exact=exact, key=key)

    def column_entry_matches(self, schema_name, table_name, column_name, exact=False, key=None):
        return self.table_entry_matches(schema_name, table_name, exact=exact, key=key) and self.matches('column',
                                                                                                        column_name,
                                                                                                        exact=exact,
                                                                                                        key=key)


class BaseSpecList:
    def __init__(self, type, dictlist=None, strict=True):
        dictlist = dictlist if dictlist else []
        self.type = type
        self.specs = []
        self.add_list(dictlist)
        self.strict = strict

    def get_specs(self):
        return self.specs

    def add_spec(self, spec):
        self.specs.append(spec)

    def add_list(self, dictlist):
        for d in dictlist:
            if len(d) > 0:
                self.add_spec(self.type(d))

    def find_best_schema_spec(self, schema_name, key=None):
        results = []
        for spec in self.specs:
            if spec.schema_entry_matches(schema_name, key=key):
                results.append(spec)
        if len(results) == 0:
            return None
        if len(results) == 1:
            return results[0]
        if len(results) > 1:
            exact_match = None
            for spec in results:
                if spec.schema_entry_matches(schema_name, exact=True, key=key):
                    if exact_match is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-match spec for schema {s}, key {k}".format(s=schema_name,
                                                                                                k=str(key)))
                        else:
                            print("WARNING: More than one exact-match spec for schema {s}, key {k}, skipping".format(
                                s=schema_name, k=str(key)))
                    else:
                        exact_match = spec
            if exact_match is None:
                if self.strict:
                    raise ValueError(
                        "More than one regexp match and no exact-match spec for schema {s}, key {k}".format(
                            s=schema_name, k=str(key)))
                else:
                    print(
                        "WARNING: More than one regexp match and no exact-match spec for schema {s}, key {k},"
                        " skipping".format(s=schema_name, k=str(key)))
            return exact_match

    def find_catalog_spec(self, key):
        results = []
        for spec in self.specs:
            if spec.get("uri") == key:
                return spec

    def find_best_foreign_key_spec(self, table_schema_name, table_name, names, key=None):
        exact_specs = []
        pattern_specs = []
        for n in names:
            new = self.find_best_foreign_key_name_spec(table_schema_name, table_name, n[0], n[1], key)
            if new is not None:
                if new.is_exact_spec():
                    exact_specs.append(new)
                else:
                    pattern_specs.append(new)
        if len(exact_specs) == 1:
            return exact_specs[0]
        elif len(exact_specs) == 0:
            if len(pattern_specs) == 1:
                return pattern_specs[0]
            elif len(pattern_specs) == 0:
                return None
            else:
                if self.strict:
                    raise ValueError(
                        "No exact-match and more than one pattern-match foreign key specification for foreign key with"
                        " names {n} on table {ts}.{t} : {s}".format(
                            n=str(names), t=table_name, ts=table_schema_name, s=str(list(pattern_specs))))
                else:
                    print(
                        "WARNING: No exact-match and more than one pattern-match foreign key specification for foreign"
                        " key with names {n} on table {ts}.{t} : {s}, skipping".format(
                            n=str(names), t=table_name, ts=table_schema_name, s=str(list(pattern_specs))))
        else:
            if self.strict:
                raise ValueError(
                    "More than one exact-match foreign key specification for foreign key with names {n} on table"
                    " {ts}.{t} : {s}".format(
                        n=str(names), t=table_name, ts=table_schema_name, s=str(list(exact_specs))))
            else:
                print(
                    "WARNING: More than one exact-match foreign key specification for foreign key with names "
                    "{n} on table {ts}.{t} : {s}, skipping".format(
                        n=str(names), t=table_name, ts=table_schema_name, s=str(list(exact_specs))))

    def find_best_foreign_key_name_spec(self, table_schema_name, table_name, fkey_schema_name, fkey_name, key):
        results = []
        for spec in self.specs:
            if spec.foreign_key_entry_matches(table_schema_name, table_name, fkey_schema_name, fkey_name, key=key):
                results.append(spec)
        if len(results) == 0:
            return None
        if len(results) == 1:
            return results[0]
        if len(results) > 1:
            result = None
            for spec in results:
                if spec.foreign_key_entry_matches(table_schema_name, table_name, fkey_schema_name, fkey_name,
                                                  exact=True, key=key):
                    if result is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-match entry for foreign key {fs}.{f} on table {s}.{t}: {r}".format(
                                    s=table_schema_name, t=table_name, fs=fkey_schema_name, f=fkey_name,
                                    r=str(results)))
                        else:
                            print(
                                "WARNING: More than one exact-match entry for foreign key {fs}.{f} on table "
                                "{s}.{t}: {r}, skipping".format(
                                    s=table_schema_name, t=table_name, fs=fkey_schema_name, f=fkey_name,
                                    r=str(results)))
                    else:
                        result = spec
            if result is None:
                if self.strict:
                    raise ValueError(
                        "More than one matching entry for foreign key {fs}.{f} on table {s}.{t} but no exact match: "
                        "{r}".format(
                            s=table_schema_name, t=table_name, fs=fkey_schema_name, f=fkey_name, r=str(results)))
                else:
                    print(
                        "WARNING: More than one matching entry for foreign key {fs}.{f} on table {s}.{t} but no exact "
                        "match: {r}, skipping".format(
                            s=table_schema_name, t=table_name, fs=fkey_schema_name, f=fkey_name, r=str(results)))
            return result

    def find_best_table_spec(self, schema_name, table_name, key=None):
        results = []
        for spec in self.specs:
            if spec.table_entry_matches(schema_name, table_name, key=key):
                results.append(spec)
        if len(results) == 0:
            return None
        if len(results) == 1:
            return results[0]
        if len(results) > 1:
            result = None
            for spec in results:
                if spec.table_entry_matches(schema_name, table_name, exact=True, key=key):
                    if result is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-match entry for table {s}.{t}".format(s=schema_name, t=table_name))
                        else:
                            print("WARNING: More than one exact-match entry for table {s}.{t}, skipping".format(
                                s=schema_name, t=table_name))
                    else:
                        result = spec
            if result is not None:
                return result
            for spec in results:
                if spec.schema_entry_matches(schema_name, exact=True, key=key):
                    if result is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-schema and no exact-table acl entry for {s}.{t}".format(
                                    s=schema_name, t=table_name))
                        else:
                            print(
                                "WARNING: More than one exact-schema and no exact-table acl entry for {s}.{t},"
                                " skipping".format(
                                    s=schema_name, t=table_name))
                    else:
                        result = spec
            return result

    def find_best_single_foreign_key_spec(self, schema_name, table_name, fkey_name, key=None):
        results = []
        for spec in self.specs:
            if spec.foreign_key_entry_matches(schema_name, table_name, fkey_name, key=key):
                results.append(spec)
        if len(results) == 0:
            return None
        if len(results) == 1:
            return results[0]
        if len(results) > 1:
            result = None
            for spec in results:
                if spec.is_exact_spec():
                    if result is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-match entry for foreign key {s}.{f}".format(s=schema_name,
                                                                                                 f=fkey_name))
                        else:
                            print(
                                "WARNING: More than one exact-match entry for foreign key {s}.{f}".format(s=schema_name,
                                                                                                          f=fkey_name))
                    else:
                        result = spec
            if result is not None:
                return result
            for spec in results:
                if spec.schema_entry_matches(schema_name, exact=True, key=key):
                    if result is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-schema and no exact-fkey entry for {s}.{f}, skipping".format(
                                    s=schema_name, f=fkey_name))
                        else:
                            print(
                                "WARNING: More than one exact-schema and no exact-fkey entry for {s}.{f},"
                                " skipping".format(
                                    s=schema_name, f=fkey_name))
                    else:
                        result = spec
            return result

    def find_best_column_spec(self, schema_name, table_name, column_name, key=None):
        results = []
        for spec in self.specs:
            if spec.column_entry_matches(schema_name, table_name, column_name, key=key):
                results.append(spec)
                #                print("column_entry_matches: {s}".format(s=str(spec)))
        if len(results) == 0:
            return None
        if len(results) == 1:
            return results[0]
        if len(results) > 1:
            result = None
            for spec in results:
                if spec.is_exact_spec():
                    if result is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-match acl entry for column {s}.{t}.{c}, key {k}".format(
                                    s=schema_name, t=table_name, c=column_name, k=str(key)))
                        else:
                            print(
                                "WARNING: More than one exact-match acl entry for column {s}.{t}.{c},"
                                " key {k}, skipping".format(
                                    s=schema_name, t=table_name, c=column_name, k=str(key)))
                    else:
                        result = spec
            if result is not None:
                return result
            for spec in results:
                if spec.table_entry_matches(schema_name, table_name, exact=True, key=key):
                    if result is not None:
                        if self.strict:
                            raise ValueError(
                                "More than one exact-schema/exact-table and no exact-column acl entry for column "
                                "{s}.{t}.{c}, key {k}".format(
                                    s=schema_name, t=table_name, c=column_name, k=str(key)))
                        else:
                            print(
                                "WARNING: More than one exact-schema/exact-table and no exact-column acl entry for "
                                "column {s}.{t}.{c}, key {k}, skipping".format(
                                    s=schema_name, t=table_name, c=column_name, k=str(key)))
                    else:
                        result = spec
            return result


class ConfigBaseCLI(BaseCLI):
    def __init__(self, description, epilog, version):
        BaseCLI.__init__(self, description, epilog, version)
        self.remove_options(['--config-file'])
        self.parser.add_argument('--config-file', metavar='<file>', help="Path to a configuration file.", required=True)
        self.parser.add_argument('-s', '--schema', help="schema name", default=None, action='append')
        self.parser.add_argument('-t', '--table', help="table name", default=None, action='append')
        self.parser.add_argument('-n', '--dryrun', help="dryrun", action="store_true")
        self.parser.add_argument('-v', '--verbose', help="verbose", action="store_true")
        self.parser.add_argument('catalog', help="catalog number", type=int)
        self.parser.set_defaults(host=platform.uname()[1])

    @classmethod
    def check_schema_table_args(cls, args):
        # Allowed combinations:
        # - multiple schemas, no tables
        # - 1 schema, 1 table
        # - no schemas, no tables
        if args.table is not None:
            if len(args.table) > 1:
                raise ValueError("More than one table specified")
            if args.schema is None or len(args.schema) != 1:
                raise ValueError("Table specified without exactly one schema")

    @classmethod
    def get_schema_arg_list(cls, args):
        cls.check_schema_table_args(args)
        if args.schema is None:
            return [None]
        else:
            return args.schema

    @classmethod
    def get_table_arg(cls, args):
        cls.check_schema_table_args(args)
        if args.table is None:
            return None
        else:
            return args.table[0]



import os
import sys
import json
import traceback
import argparse
from deriva.core import BaseCLI, KeyValuePairArgs, format_credential, format_exception, urlparse
from deriva.transfer import DerivaBackup, DerivaDownloadCLI


class DerivaBackupCLI(DerivaDownloadCLI):
    def __init__(self, description, epilog, **kwargs):

        DerivaDownloadCLI.__init__(self, description, epilog, **kwargs)
        mutex_group = self.parser.add_mutually_exclusive_group()
        mutex_group.add_argument("--no-data", action="store_true",
                                 help="Do not export data, export schema only.")
        mutex_group.add_argument("--no-schema", action="store_true",
                                 help="Do not export schema, export data only.")
        bag_mutex_group = self.parser.add_mutually_exclusive_group()
        bag_mutex_group.add_argument("--no-bag", action="store_true",
                                     help="Do not store the output in a bag container.")
        bag_mutex_group.add_argument("--include-assets", choices=['full', 'references'],
                                     help="Include related file assets in output bag. Use \"full\" to download "
                                          "related assets to the output bag. Use \"references\" to store "
                                          "references to asset files in the bag's \"fetch.txt\" file.")
        self.parser.add_argument("--bag-archiver", choices=['zip', 'tgz', 'bz2'],
                                 help="Format for compressed bag output.")
        self.parser.add_argument("--exclude-data", default=list(),
                                 type=lambda s: [item.strip() for item in s.split(',')],
                                 metavar="<schema>, <schema:table>, ...",
                                 help="List of comma-delimited schema-name and/or schema-name/table-name to "
                                      "exclude from data export, in the form <schema> or <schema:table>.")

    @classmethod
    def get_downloader(cls, *args, **kwargs):
        return DerivaBackup(*args, dcctx_cid="cli/" + DerivaBackupCLI.__name__, **kwargs)



import os
import sys
import json
import traceback
import argparse
import requests
from requests.exceptions import HTTPError, ConnectionError
from deriva.transfer import GenericDownloader
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError, \
    DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError
from deriva.core import BaseCLI, KeyValuePairArgs, format_credential, format_exception, urlparse


class DerivaDownloadCLI(BaseCLI):
    def __init__(self, description, epilog, **kwargs):

        BaseCLI.__init__(self, description, epilog, **kwargs)
        self.parser.add_argument("--catalog", default=1, metavar="<1>", help="Catalog number. Default: 1")
        self.parser.add_argument("output_dir", metavar="<output dir>", help="Path to an output directory.")
        self.parser.add_argument("envars", metavar="[key=value key=value ...]",
                                 nargs=argparse.REMAINDER, action=KeyValuePairArgs, default={},
                                 help="Variable length of whitespace-delimited key=value pair arguments used for "
                                      "string interpolation in specific parts of the configuration file. "
                                      "For example: key1=value1 key2=value2")

    @classmethod
    def get_downloader(cls, *args, **kwargs):
        return GenericDownloader(*args, dcctx_cid="cli/" + DerivaDownloadCLI.__name__, **kwargs)

    @classmethod
    def download(cls, args):

        assert args.host, "A hostname is required!"
        server = dict()
        server["catalog_id"] = args.catalog
        if args.host.startswith("http"):
            url = urlparse(args.host)
            server["protocol"] = url.scheme
            server["host"] = url.netloc
        else:
            server["protocol"] = "https"
            server["host"] = args.host

        downloader = cls.get_downloader(server, **vars(args))

        return downloader.download()

    def main(self):
        try:
            args = self.parse_cli()
        except ValueError as e:
            sys.stderr.write(str(e))
            return 2
        if not args.quiet:
            sys.stderr.write("\n")

        try:
            try:
                downloaded = self.download(args)
                sys.stdout.write("\n%s\n" % (json.dumps(downloaded)))
            except ConnectionError as e:
                raise DerivaDownloadError("Connection error occurred. %s" % format_exception(e))
            except HTTPError as e:
                if e.response.status_code == requests.codes.unauthorized:
                    raise DerivaDownloadAuthenticationError(
                        "The requested service requires authentication and a valid login session could "
                        "not be found for the specified host. Server responded: %s" % e)
                elif e.response.status_code == requests.codes.forbidden:
                    raise DerivaDownloadAuthorizationError(
                        "A requested operation was forbidden. Server responded: %s" % e)
        except (DerivaDownloadError, DerivaDownloadConfigurationError,
                DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError) as e:
            sys.stderr.write(("\n" if not args.quiet else "") + format_exception(e))
            if args.debug:
                traceback.print_exc()
            return 1
        except:
            sys.stderr.write("An unexpected error occurred.")
            traceback.print_exc()
            return 1
        finally:
            if not args.quiet:
                sys.stderr.write("\n\n")
        return 0




import uuid
import sys
import os
import json
import requests
from multiprocessing import Queue
from . import get_new_requests_session, urlquote_dcctx, ConcurrentUpdate, NotModified, DEFAULT_HEADERS, DEFAULT_SESSION_CONFIG


class DerivaClientContext (dict):
    """Represent Deriva-Client-Context header content.

    Well-known keys (originally defined for Chaise):

    - `cid`: client application ID i.e. program name

    - `wid`: window ID i.e. request stream ID

    - `pid`: page ID i.e. sub-stream ID

    - `action`: UX action embodied by request(s)

    - `table`: table upon which application is focused (if any)

    - `uinit`: True if request initiated by user action

    Default values to use process-wide:

    - `cid`:code:: os.path.basename(sys.argv[0]) if available

    - `wid`:code:: a random UUID

    The process-wide defaults MAY be customized by mutating
    `DerivaClientContext.defaults`:code: prior to constructing instances.

    """
    defaults = {
        'cid': os.path.basename(sys.argv[0]) if len(sys.argv) > 0 and sys.argv[0] else None,
        'wid': uuid.uuid4().hex,
    }

    def __init__(self, *args, **kwargs):
        """Initialize DerivaClientContext from keywords or dict-like.

           Class-wide defaults are used for fields not set explicitly during construction.
        """
        super(DerivaClientContext, self).__init__(*args, **kwargs)
        self.set_defaults()
        self.prune()

    def set_defaults(self, defaults=None):
        """Set default key-values in self if key not already set."""
        if defaults is None:
            defaults = self.defaults
        for k, v in defaults.items():
            if v is not None and (k not in self or self[k] is None):
                self[k] = v

    def prune(self):
        """Prune redundant keys to shorten context representation.

           Keys with None value or uinit=False are unnecessary as these
           are implicitly assumed when absent.

        """
        for k, v in self.items():
            if v is None:
                del self[k]
        if 'uinit' in self and not self['uinit']:
            del self['uinit']

    def encoded(self):
        """Encode self as string suitable for Deriva-Client-Context HTTP header."""
        self.set_defaults()
        self.prune()
        x = urlquote_dcctx(json.dumps(self, indent=None, separators=(',', ':')))
        return x

    def merged(self, overrides):
        c = DerivaClientContext(self)
        c.update(overrides)
        return c

class DerivaPathError (ValueError):
    pass

def _response_raise_for_status(self):
    """Raises requests.HTTPError if status code indicates an error.

    This unbound method can be monkey-patched onto a requests.Response
    instance or manually invoked on one.

    """
    if 400 <= self.status_code < 600:
        details = " Details: %s"
        raise requests.HTTPError(
            u'%s %s Error: %s for url: [%s]%s' % (
                self.status_code,
                'Client' if self.status_code < 500 else 'Server',
                self.reason,
                self.url,
                details % self.content if self.content else "",
            ),
            response=self
        )


class DerivaBinding (object):
    """This is a base-class for implementation purposes. Not useful for clients."""

    def __init__(self, scheme, server, credentials=None, caching=True, session_config=None):
        """Create HTTP(S) server binding.

           Arguments:
             scheme: 'http' or 'https'
             server: server FQDN string
             credentials: credential secrets, e.g. cookie
             caching: whether to retain a GET response cache

           Deriva Client Context: You MAY mutate self.dcctx to
           customize the context for this service endpoint prior to
           invoking web requests.  E.g.:

             self.dcctx['cid'] = 'my application name'

           You MAY also supply custom per-request context by passing a
           headers dict to web request methods, e.g. 

             self.get(..., headers={'deriva-client-context': {'action': 'myapp/function1'}})

           This custom header will be merged as override values with
           the default context in self.dcctx in order to form the
           complete context for the request.

        """
        self._base_server_uri = "%s://%s" % (
            scheme,
            server
        )
        self._server_uri = self._base_server_uri

        if not session_config:
            session_config = DEFAULT_SESSION_CONFIG
        self._session = None
        self._get_new_session(session_config)

        self._caching = caching
        self._cache = {}

        self._response_raise_for_status = _response_raise_for_status

        self.dcctx = DerivaClientContext()

        self.set_credentials(credentials, server)

    def get_server_uri(self):
        return self._server_uri

    def _get_new_session(self, session_config):
        self._close_session()
        self._session = get_new_requests_session(self._server_uri + '/', session_config)

    def _pre_get(self, path, headers):
        self.check_path(path)
        url = self._server_uri + path
        headers = headers.copy()
        prev_response = self._cache.get(url)
        if prev_response and 'etag' in prev_response.headers \
           and not ('if-none-match' in headers or 'if-match' in headers):
            headers['if-none-match'] = prev_response.headers['etag']
        else:
            prev_response = None
        headers['deriva-client-context'] = self.dcctx.merged(headers.get('deriva-client-context', {})).encoded()
        return url, headers, prev_response

    def _pre_mutate(self, path, headers, guard_response=None):
        self.check_path(path)
        url = self._server_uri + path
        headers = headers.copy()
        if guard_response and 'etag' in guard_response.headers:
            headers['If-Match'] = guard_response.headers['etag']
        headers['deriva-client-context'] = self.dcctx.merged(headers.get('deriva-client-context', {})).encoded()
        return url, headers

    @staticmethod
    def check_path(path):
        if not path:
            raise DerivaPathError("Path not specified")

        if not path.startswith("/"):
            raise DerivaPathError("Malformed path error (not rooted with \"/\"): %s" % path)

    @staticmethod
    def _raise_for_status_304(r, p, raise_not_modified):
        if r.status_code == 304:
            if raise_not_modified:
                raise NotModified(p or r)
            else:
                return p or r

        _response_raise_for_status(r)
        setattr(r, 'raise_for_status', _response_raise_for_status.__get__(r))
        return r

    @staticmethod
    def _raise_for_status_412(r):
        if r.status_code == 412:
            raise ConcurrentUpdate(r)
        _response_raise_for_status(r)
        setattr(r, 'raise_for_status', _response_raise_for_status.__get__(r))
        return r

    def set_credentials(self, credentials, server):
        if not credentials:
            return
        assert self._session is not None
        if 'cookie' in credentials:
            cname, cval = credentials['cookie'].split('=', 1)
            self._session.cookies.set(cname, cval, domain=server, path='/')
        elif 'bearer-token' in credentials:
            self._session.headers.update({'Authorization': 'Bearer {token}'.format(token=credentials['bearer-token'])})
        elif 'username' in credentials and 'password' in credentials:
            self.post_authn_session(credentials)

    def get_authn_session(self):
        headers = { 'deriva-client-context': self.dcctx.encoded() }
        r = self._session.get(self._base_server_uri + "/authn/session", headers=headers)
        _response_raise_for_status(r)
        return r

    def post_authn_session(self, credentials):
        headers = { 'deriva-client-context': self.dcctx.encoded() }
        r = self._session.post(self._base_server_uri + "/authn/session", data=credentials, headers=headers)
        _response_raise_for_status(r)
        return r

    def head(self, path, headers=DEFAULT_HEADERS, raise_not_modified=False):
        """Perform HEAD request, returning response object.

           Arguments:
             path: the path within this bound server
             headers: headers to set in request
             raise_not_modified: raise HTTPError for 304 response
               status when true.

           May consult built-in cache and apply 'if-none-match'
           request header unless input headers already include
           'if-none-match' or 'if-match'. On cache hit, returns cached
           response unless raise_not_modified=true. Cached response
           may include content retrieved by GET on the same resource.

        """
        url, headers, prev_response = self._pre_get(path, headers)
        return self._raise_for_status_304(
            self._session.head(url, headers=headers),
            prev_response,
            raise_not_modified
        )
        
    def get(self, path, headers=DEFAULT_HEADERS, raise_not_modified=False, stream=False):
        """Perform GET request, returning response object.

           Arguments:
             path: the path within this bound server
             headers: headers to set in request
             raise_not_modified: raise HTTPError for 304 response
               status when true.
             stream: whether to defer content retrieval to 
               streaming access mode on response object.

           May consult built-in cache and apply 'if-none-match'
           request header unless input headers already include
           'if-none-match' or 'if-match'. On cache hit, returns cached
           response unless raise_not_modified=true.

           Caching of new results is disabled when stream=True.

        """
        if headers is None:
            headers = {}
        url, headers, prev_response = self._pre_get(path, headers)
        r = self._raise_for_status_304(
            self._session.get(url, headers=headers),
            prev_response,
            raise_not_modified
        )
        if self._caching and not stream:
            self._cache[url] = r
        return r

    def post(self, path, data=None, json=None, headers=DEFAULT_HEADERS):
        """Perform POST request, returning response object.

           Arguments:
             path: the path within this bound server
             data: a buffer or file-like content value
             json: data to serialize as JSON content
             headers: headers to set in request

           Raises ConcurrentUpdate for 412 status.

        """
        url, headers = self._pre_mutate(path, headers)
        r = self._session.post(url, data=data, json=json, headers=headers)
        return self._raise_for_status_412(r)

    def put(self, path, data=None, json=None, headers=DEFAULT_HEADERS, guard_response=None):
        """Perform PUT request, returning response object.

           Arguments:
             path: the path within this bound server
             data: a buffer or file-like content value
             json: data to serialize as JSON content
             headers: headers to set in request
             guard_response: expected current resource state
               as previously seen response object.

           Uses guard_response to build appropriate 'if-match' header
           to assure change is only applied to expected state.

           Raises ConcurrentUpdate for 412 status.

        """ 
        url, headers = self._pre_mutate(path, headers, guard_response)
        r = self._session.put(url, data=data, json=json, headers=headers)
        return self._raise_for_status_412(r)
   
    def delete(self, path, headers=DEFAULT_HEADERS, guard_response=None):
        """Perform DELETE request, returning response object.

           Arguments:
             path: the path within this bound server
             headers: headers to set in request
             guard_response: expected current resource state
               as previously seen response object.

           Uses guard_response to build appropriate 'if-match' header
           to assure change is only applied to expected state.

           Raises ConcurrentUpdate for 412 status.

        """
        url, headers = self._pre_mutate(path, headers, guard_response)
        r = self._session.delete(url, headers=headers)
        return self._raise_for_status_412(r)

    def _close_session(self):
        if self._session is not None:
            self._session.close()

    def __del__(self):
        self._close_session()



import os
import json
from collections import OrderedDict
from deriva.transfer.download.processors.transform.base_transform_processor import BaseTransformProcessor


class FastaExportTransformProcessor(BaseTransformProcessor):
    def __init__(self, envars=None, **kwargs):
        super(FastaExportTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()

    def process(self):
        self.convert_json_file_to_fasta(self.input_abspath, self.output_abspath, self.processor_params)
        return super(FastaExportTransformProcessor, self).process()

    @staticmethod
    def convert_json_obj_to_fasta_lines(obj, params):
        column_map = None
        comment_line = ''
        data_line = ''
        if params:
            column_map = params.get("column_map", None)

        if (len(obj) > 1) and (not column_map):
            raise RuntimeError("FASTA converter input data has more than one element and no column map was specified.")
        elif len(obj) == 1:
            k, v = obj.popitem()
            lines = str("%s\n" % v)
        else:
            for k, v in obj.items():
                if column_map.get(k, None) == "comment":
                    if comment_line:
                        comment_line += str(" | %s" % v)
                    else:
                        comment_line = str("> %s" % v)
                elif column_map.get(k, None) == "data":
                    data_line += v
            lines = str("%s\n%s\n" % (comment_line, data_line))

        return lines

    @staticmethod
    def convert_json_file_to_fasta(input_file, output_file, params):
        with open(input_file, "r") as input_data, open(output_file, "w") as output_data:
            line = input_data.readline().lstrip()
            input_data.seek(0)
            is_json_stream = False
            if line.startswith('{'):
                data = input_data
                is_json_stream = True
            else:
                try:
                    data = json.load(input_data, object_pairs_hook=OrderedDict)
                except ValueError:
                    data = {}

            for entry in data:
                if is_json_stream:
                    entry = json.loads(entry, object_pairs_hook=OrderedDict)

                lines = convert_json_obj_to_fasta_lines(entry, params)
                output_data.writelines(lines)
